from ...problem_operations.conflicts.blocked_magnet import conflicted_magnet
from ...general_operations.geometry_intersections.circle_with_rectangle import circle_rectangle_intersection
from ...general_operations.geometry_intersections.rectangle_with_rectangle import rectangle_rectangle_intersection
from ...hector.magnets.circular import is_circular_magnet
from ...hector.magnets.rectangular import is_rectangular_magnet


def circle_blocking_rectangle(circle,rectangle):

    all_blocked_pickup_areas = []

    for pickup_area in rectangle.pickup_areas:

        points = circle_rectangle_intersection(circle, pickup_area)

        if points:

            # plt.figure()
            # circle.draw_circle('r')
            # plt.text(circle.center[0], circle.center[1],str(int(circle.index)), fontsize=6)
            #
            # rectangle.draw_rectangle('c')
            # plt.text(rectangle.center[0], rectangle.center[1],str(int(rectangle.index)), fontsize=6)
            #
            # pickup_area.draw_rectangle('r')

            # for point in points:
            #    plt.plot(point[0],point[1],'or')

            all_blocked_pickup_areas.append(conflicted_magnet(circle, rectangle, pickup_area))


    return all_blocked_pickup_areas

def rectangle_blocking_circle(circle,rectangle):

    all_blocked_pickup_areas = []

    for pickup_area in circle.pickup_areas:

        points = rectangle_rectangle_intersection(rectangle, pickup_area)

        if points:

            # plt.figure()
            # circle.draw_circle('r')
            # plt.text(circle.center[0], circle.center[1],str(int(circle.index)), fontsize=6)
            #
            # rectangle.draw_rectangle('c')
            # plt.text(rectangle.center[0], rectangle.center[1],str(int(rectangle.index)), fontsize=6)
            #
            # pickup_area.draw_rectangle('r')

            # for point in points:
            #   plt.plot(point[0],point[1],'or')

            all_blocked_pickup_areas.append(conflicted_magnet(rectangle, circle, pickup_area))

    return all_blocked_pickup_areas

def check_conflict_circle_rectangle_magnets(magnet_pair):

    all_blocked_pickup_areas = []

    for magnet_1, magnet_2 in zip([0, 1], [1, 0]):

        if is_circular_magnet(magnet_pair[magnet_1]):

            blocked_pickup_area = circle_blocking_rectangle(magnet_pair[magnet_1], magnet_pair[magnet_2])

            if blocked_pickup_area:

                all_blocked_pickup_areas.extend(blocked_pickup_area)


        elif is_rectangular_magnet(magnet_pair[magnet_1]):

            blocked_pickup_area = rectangle_blocking_circle(magnet_pair[magnet_2], magnet_pair[magnet_1])

            if blocked_pickup_area:

                all_blocked_pickup_areas.extend(blocked_pickup_area)

    return all_blocked_pickup_areas



from ...hector.magnets.circular import is_circular_magnet
from ...hector.magnets.rectangular import is_rectangular_magnet

def circle_and_circle_magnets(pair):

    return is_circular_magnet(pair[0]) and is_circular_magnet(pair[1])

def circle_rectangle_magnets(pair):

    if is_circular_magnet(pair[0]) and is_rectangular_magnet(pair[1]):
        return True

    elif is_rectangular_magnet(pair[0]) and is_circular_magnet(pair[1]):
        return True

    else:
        return False

def rectangle_rectangle_magnets(pair):

    return is_rectangular_magnet(pair[0]) and is_rectangular_magnet(pair[1])


import matplotlib.pyplot as plt

class segment():

    def __init__(self,start_point,end_point):
        self.start_point = start_point
        self.end_point = end_point

    def create_line_equation(self):

        slope = (self.end_point[1] - self.start_point[1]) / (self.end_point[0] - self.start_point[0])
        origin = self.end_point[1] - slope * self.end_point[0]
        return line(slope, origin)

    def vertical(self):

        return self.start_point[0] == self.end_point[0]

    def point_is_on_segment(self,point):

        if min(self.start_point[0], self.end_point[0]) <= point[0] <= max(self.start_point[0],self.end_point[0]):
            if min(self.start_point[1], self.end_point[1]) <= point[1] <= max(self.start_point[1],self.end_point[1]):

                return True
            else:
                return False
        else:
            return False

    def vector(self):

        return [self.end_point[0] - self.start_point[0],
                self.end_point[1] - self.start_point[1]]

    def draw_segment(self,colour):

        plt.plot([self.start_point[0], self.end_point[0]], [self.start_point[1], self.end_point[1]], colour)
        plt.axis('scaled')

class line():

    def __init__(self,slope,origin):
        self.slope = slope
        self.origin = origin

    def calculate_y_from_x(self, x):
        return self.slope * x + self.origin




from ...general_operations.geometry_shapes.rectangle import rectangle
from ...hector.constants import rectangle_magnet_length,rectangle_magnet_width,robot_arm_width
from math import sin, cos, pi
from ...hector.magnets.pickup_areas import inward, outward

class rectangular_magnet(rectangle):

    def __init__(self,center,orientation,index,galaxyORstar,Re,mu_1re,Mstar,magnet_label,hexabundle,rads,rotation_pickup,rotation_putdown,azAngs,rectangular_magnet_input_orientation,IDs):
        length = rectangle_magnet_length
        width  = rectangle_magnet_width
        super().__init__(center,length,width,orientation)
        self.index = index
        self.placement_index = 0
        self.galaxyORstar = galaxyORstar
        self.Re = Re
        self.mu_1re = mu_1re
        self.Mstar = Mstar
        self.magnet_label = magnet_label
        self.hexabundle = hexabundle
        self.rads = rads
        self.rotation_pickup = rotation_pickup
        self.rotation_putdown = rotation_putdown
        self.azAngs = azAngs
        self.rectangular_magnet_input_orientation = rectangular_magnet_input_orientation
        self.IDs = IDs

    def calculate_center_magnet_to_center_pickup_area_length(self):

        # center_magnet_to_center_pickup_area_length = 0.25 * (self.length + robot_arm_width)  # Tiphaine's version

        # if (robot_arm_width < ((rectangle_magnet_length - robot_arm_width) / 2)):
        #     center_magnet_to_center_pickup_area_length = 0.25 * (rectangle_magnet_length - robot_arm_width)
        # elif (robot_arm_width >= ((rectangle_magnet_length - robot_arm_width) / 2)):
        #     center_magnet_to_center_pickup_area_length = robot_arm_width / 2

        center_magnet_to_center_pickup_area_length = 0.25 * (10 + robot_arm_width)  # generic one for all RAW sizes

        return center_magnet_to_center_pickup_area_length

    def calculate_center_coordinate_outward_pickuparea(self):

        center_coordinates = \
        [self.center[0] + self.calculate_center_magnet_to_center_pickup_area_length() * sin((pi / 180.0) * (self.orientation)),
         self.center[1] + self.calculate_center_magnet_to_center_pickup_area_length() * cos((pi / 180.0) * (self.orientation))]

        return center_coordinates

    def calculate_center_coordinate_inward_pickuparea(self):

        center_coordinates = \
        [self.center[0] - self.calculate_center_magnet_to_center_pickup_area_length() * sin((pi / 180.0) * (self.orientation)),
         self.center[1] - self.calculate_center_magnet_to_center_pickup_area_length() * cos((pi / 180.0) * (self.orientation))]

        return center_coordinates

    def create_pickup_areas(self):

        self.pickup_areas = \
        [inward(self.calculate_center_coordinate_inward_pickuparea(), self.orientation),
         outward(self.calculate_center_coordinate_outward_pickuparea(),self.orientation)]

        return self.pickup_areas

def is_rectangular_magnet(magnet):
    return isinstance(magnet, rectangular_magnet)



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import yaml
import bisect
# For reading fits tables
from ..misc import pandas_tools as P 
import os

from ..misc import misc_tools as misc

# def _load_config(SourceCat):

#     with open(f"Config/{SourceCat}.yml", 'r') as ymlfile:
#         config = yaml.safe_load(ymlfile)

#     return config


# def _read_table(fname):
#     """
#     Read in a table, either with fits or pandas
#     """
#     # Read in the whole table. May be an issue with massive tables?
#     extension = fname.split('.')[-1]
#     if extension == 'csv':
#         table = pd.read_csv(fname)
#     elif extension == 'tsv':
#         table = pd.read_csv(fname, sep='\t')
#     elif extension == 'fits':
#         table = P.load_FITS_table_in_pandas(fname)
#         #raise TypeError("Not yet written the code for fits... Exiting!")
#     else:
#         raise TypeError(f"Catalogue {fname} data type \
#             not understood")
#     return table


def load_table(config, quiet=False):
    """
    Read in the table, apply the selection cuts and make our cut-down table
    """

    table = misc._read_table(config['input_master_catalogue_filename'])

    selection = config['selection']
    if not quiet:
        print(f"Selected rows with: {selection}")

    # Cut down our table based on the python code in the config file
    reduced_table = table[eval(selection)]

    # Get the names of each column we want and their index in this
    # specific table
    column_names = np.array(list(config['Columns'].keys()))
    column_indices = np.array(list(config['Columns'].values()))

    # Find the columns we want which exist in this table- i.e. those which
    # have an index which isn't "None"
    if np.any(column_indices == 'None'):
        not_none_mask = column_indices != 'None'
    else:
        not_none_mask = np.ones_like(column_indices, dtype=bool)

    # Columns which we'll select from the table
    columns_to_select = column_indices[not_none_mask]
    assert np.unique(columns_to_select).size == columns_to_select.size, \
        'Looks like we have a repeated index in the config file'

    # ...and the names we'll rename them
    column_names_for_renaming = column_names[not_none_mask]

    # final table- select our columns
    final_table = reduced_table.iloc[:, columns_to_select.astype(int)]

    # Rename the columns- this is a dictionary of old_name : new_name
    renamed_columns = dict(zip(final_table.columns.tolist(),
                               column_names_for_renaming))

    final_table = final_table.rename(index=str, columns=renamed_columns)
    if not quiet:
        print("Renaming Columns:")
        for old_name, new_name in renamed_columns.items():
            print(f"\t {old_name} --> {new_name}")
            # Check everything which should be a float actually is
            if new_name in ['Mstar', 'z', 'Re', 'GAL_MAG_R', 'GAL_MAG_I']:
                final_table[f'{new_name}'] = pd.to_numeric(
                    final_table[f'{new_name}'], errors='coerce')

    # Now add in the columns which didn't exist in the table as nans
    none_values = np.where(column_indices == 'None')[0]
    for index in none_values:
        column_name = column_names[index]
        final_table[column_name] = np.nan

    return final_table



def sparse_select_flat_in_Mass(table, bins, MStarSparseCutBin):
    """
    Take a table of stellar masses and bin into 5 bins between MstarMin and
    MstarMax.
    Take the number in the MSparseCut1 bin (call it n) and then flatten all
    bins to have at most n galaxies in
    """

    # Get the bin which each value should be in
    bin_indices = np.digitize(table.Mstar, bins=bins)
    numMstar, edges = np.histogram(table.Mstar, bins=bins)

    Num_per_Mstarbin = numMstar[MStarSparseCutBin]

    # Now loop through each bin and select which bin members to keep
    # import ipdb; ipdb.set_trace()
    # indices_to_keep = []
    # for bin_num in np.unique(bin_indices):
    #     indices_in_bin = np.where(bin_indices == bin_num)[0]
    #     N_in_bin = indices_in_bin.size
    #     if N_in_bin <= Num_per_Mstarbin:
    #         indices_to_keep += (indices_in_bin.tolist())
    #     else:
    #         random_choice = np.random.choice(indices_in_bin, Num_per_Mstarbin, replace=False)
    #         indices_to_keep += random_choice.tolist()

    indices_to_keep = []
    for bin_num in np.unique(bin_indices):
        indices_in_bin = np.where(bin_indices == bin_num)[0]
        N_in_bin = indices_in_bin.size
        if bin_num <= MStarSparseCutBin:
            indices_to_keep += (indices_in_bin.tolist())
        else:
            if N_in_bin <= Num_per_Mstarbin:
                indices_to_keep += (indices_in_bin.tolist())
            else:
                random_choice = np.random.choice(indices_in_bin, Num_per_Mstarbin, replace=False)
                indices_to_keep += random_choice.tolist()

    indices_to_keep = np.sort(np.array(indices_to_keep))

    return table.iloc[indices_to_keep, :]


def safe_div(x, y):
    if y == 0:
        return 0
    else:
        return x / y


def add_boxes_to_Mstar_z(fig, ax, table_data):
    """
    Taken from original code, lines 1169 to 1204
    """

    zdivisions = (0.0, 0.02, 0.04, 0.06, 0.08, 0.1)
    Mstardivisions = (9.0, 9.4, 10.3, 10.9, 12)
    Frac_Re_Cut = 2.0
    for i in [0, 1, 2, 3, 4]:
        for j in [0, 1, 2, 3]:
            ax.plot([zdivisions[i], zdivisions[i + 1]], [Mstardivisions[j], Mstardivisions[j]], '-.', color='black')
            # ax.plot([zdivisions[i],zdivisions[i] ], [pz_max(zdivisions[i]),12], '-.', color='black')
            ax.plot([zdivisions[i], zdivisions[i]], [9, 12], '-.', color='black')
            # plt.plot([0.0,0.02 ], [9.0,9.0], '.', color='black')
            # plt.plot([0.0,0.04 ], [9.4,9.4], '.', color='black')
            # plt.plot([0.0,0.1 ], [10.3,10.3], '.', color='black')
            # plt.plot([0.0,0.1 ], [10.9,10.9], '.', color='black')
            # plt.plot([0.02,0.02 ], [pz_max(0.02),12], '.', color='black')
            # plt.plot([0.04,0.04 ], [pz_max(0.04),12], '.', color='black')
            # plt.plot([0.06,0.06 ], [pz_max(0.06),12], '.', color='black')
            # plt.plot([0.08,0.08 ], [pz_max(0.08),12], '.', color='black')
            # plt.plot([0.1,0.1 ], [pz_max(0.1),12], '.', color='black')
            Num_per_boxarray = table_data[np.logical_and.reduce((table_data.z > zdivisions[i], table_data.z < zdivisions[(i + 1)], table_data.Mstar > Mstardivisions[j], table_data.Mstar < Mstardivisions[(j + 1)]))]
            # print 'Num_per_boxarray[:,Re_col]', Num_per_boxarray[:,Re_col]
            # print 'np.shape(Num_per_boxarray)', np.shape(Num_per_boxarray)
            # print 'np.shape(Num_per_boxarray[:,Re_col])', np.shape(Num_per_boxarray[:,Re_col])
            Num_per_boxtemp = Num_per_boxarray.Re
            Num_per_box = Num_per_boxtemp.size
            # Num_per_box =Num_per_boxtemp[0]
            # print 'Num_per_boxtemp, Num_per_box, Num_per_box[0]', Num_per_boxtemp, Num_per_box, Num_per_box[0]
            Num_per_box_reGT1pt5 = np.shape(Num_per_boxtemp[(Num_per_boxtemp > Frac_Re_Cut)])  # Not sure why this is two and not our cut from the command line?

            # print 'Num_per_boxtemp[(Num_per_boxtemp>Frac_Re_Cut)]',Num_per_boxtemp[(Num_per_boxtemp>Frac_Re_Cut)]
            # print 'Num_per_box_reGT1pt5', Num_per_box_reGT1pt5
            FracGTRe1pt5 = int(100 * (safe_div((float(Num_per_box_reGT1pt5[0])), (float(Num_per_box)))))
            # print 'Num_per_boxarray, Num_per_box[0], Num_per_box_reGT1pt5[0],FracGTRe1pt5', Num_per_boxarray, Num_per_box[0], Num_per_box_reGT1pt5[0],FracGTRe1pt5
            # print '(zdivisions[(i+1)]+zdivisions[(i)])/2., (Mstardivisions[(j+1)]+Mstardivisions[(j)])/2.', (zdivisions[(i+1)]+zdivisions[(i)])/2., (Mstardivisions[(j+1)]+Mstardivisions[(j)])/2.
            zboxcentre = ((zdivisions[(i + 1)] + zdivisions[(i)]) / 2.)
            Mboxcentre = ((Mstardivisions[(j + 1)] + Mstardivisions[(j)]) / 2.)
            ax.text(zboxcentre, Mboxcentre, str(FracGTRe1pt5), color='black', horizontalalignment='center', verticalalignment='center')

    return fig, ax


def calculate_SB_at_R(Sersic_index, Mu_e, multiple_of_Re):
    """
    Calculate a galaxy's surface brightness at a given multiple of Re from its surface brightness at Re

    """

    # Sersic indices
    n_table = np.arange(1, 11)
    # Values of b from the Sersic Formula
    bvalue_table = np.array([1.6783470, 3.6720608, 5.6701554, 7.6692495, 9.6687149, 11.668363, 13.667757, 
        15.667704, 17.667636, 19.667567])
    b = np.interp(Sersic_index, n_table, bvalue_table)

    # Equation from https://ned.ipac.caltech.edu/level5/March05/Graham/Graham2.html
    SB_at_R = Mu_e + 2.5 * (b / np.log(10)) * ((multiple_of_Re)**(1. / Sersic_index) - 1)
    
    return SB_at_R


class HectorSim():
    """
    A class to simulate obsverations for Hector. 

    We load in a catalogue, select the columns we want, then make plots with (in general) three samples:
        * all objects above some limiting surface brightness (plotted in RED)
        * all objects given some selection function in mass  (plotted in BLUE)
        * a subsample of the blue objects, given some sampling technique (plotted in MAGENTA)
    """

    def __init__(self, entire_table, BoundaryType, zlimit, MstarMin, MstarMax, SparseFunction, SourceCat, MSparseCut1, minRe, total_area=np.nan, quiet=False, Dec_Min=np.nan, Dec_Max=np.nan):
        """
        Args:
            BoundaryType (str): This controls both the type of selection done in stellar mass and the limiting surface brightness of the sample. Must be one of "Taipan" or "SAMI" (case insensitive). 

            zlimit (float): The maximum z to which we select galaxies. All objects at z > zlimit will be ignored in our BLUE and MAGENTA samples.

            MstarMin (float): Used to set the leftmost edge of the smallest mass bin in the mass histograms. 

            MstarMax (float): Used to set the rightmost edge of the largest mass bin in the mass histograms. 

            SparseFunction (int): Method to go from our BLUE sample to our MAGENTA sample. Must be 1 or 2. Method 1 selects MAGENTA galaxies using the following method- all galaxies below MSparseCut1 are selected. For galaxies above MSparseCut1, we take a random sample in mass such that each bin contains the same number of galaxies as in the MSparseCut1 bin. Method 2 is the same as method 1 except we upweight star-forming objects (those on the blue cloud in colour-magnitude space). 

            SourceCat (str): Name of the input catalogue to use. Must have a file names <SourceCat>.yml in the Config/ directory, which we will use to read in the catalogue. 

            MSparseCut1 (float): The bin at which we start our sparse sampling to flatten off the mass distribution. 

            minRe (float): A cut applied on the Re of targets. Can be 0.0 to not select on Re

            quiet (bool, optional): Suppress print statements. 
        """

        # Some checking on the Boundary Type
        if BoundaryType in ['taipan', "Taipan", "TAIPAN"]:
            self.BoundaryType = "Taipan"
        elif BoundaryType in ['sami', "Sami", "SAMI"]:
            self.BoundaryType = "SAMI"
        else:
            if os.path.exists(f'{BoundaryType}'):
                self.BoundaryType = 'General'
                self.BoundaryFile = f'{BoundaryType}'
            else:
                raise NameError(f"Bounday Type {BoundaryType} not understood or file does not exist")

        # Save our input parameters as class attributes
        self.zlimit = zlimit
        self.MstarMin = MstarMin
        self.MstarMax = MstarMax
        self.SparseFunction = SparseFunction
        self.MSparseCut1 = MSparseCut1
        self.minRe = minRe
        self.SourceCat = SourceCat
        self.quiet = quiet
        self.entire_table = entire_table
        self.total_area = total_area

        if not self.quiet: 
            print(f"Maximum redshift is {self.zlimit}")
            print(f"Source Catalogue is {self.SourceCat}")
            print(f"Boundary Type is {self.BoundaryType}")
            print(f"Sparse Function is type {self.SparseFunction}")
            print(f"MSparse Cut is {self.MSparseCut1}")
            print(f"Mstar is between {self.MstarMin} and {self.MstarMax}")
            print(f"Selecting objects with Re greater than {self.minRe}")

        # Check if we have HI information and update an HI_info attribute if we do
        self.HI_info = False
        if not np.all(np.isnan(self.entire_table['WALLABYflag'])) or np.all(np.isnan(self.entire_table['Dingoflag'])):
            self.HI_info = True

        # Colourmap for plotting. Should add this as an option...
        self.cm = plt.cm.get_cmap('RdYlBu_r')

        # Make a mask to get rid of obviously silly numbers
        # We'll add a check on the surface brightness values too later if we can create them
        good_data_mask = (self.entire_table.Mstar > 7) & (self.entire_table.Mstar < self.MstarMax) & (self.entire_table.Re > 0) & (self.entire_table.Re < 100)
        # Add in columns we'll need, such as g-i colour and the IFU diameter to get to 2Re
        self.entire_table['g_m_i'] = self.entire_table.loc[:, 'GAL_MAG_G'] - self.entire_table.loc[:, 'GAL_MAG_I']
        self.entire_table['IFU_diam_2Re'] = 4 * self.entire_table.loc[:, 'Re']

        # Work out surface brightness at 2 and 3 Re if we can
        if not np.all(np.isnan(self.entire_table.GAL_MU_E_R)) | np.all(np.isnan(self.entire_table.SersicIndex_r)):
            self.entire_table['GAL_MU_R_at_2Re'] = calculate_SB_at_R(self.entire_table.SersicIndex_r, self.entire_table.GAL_MU_E_R, 2)
            self.entire_table['GAL_MU_R_at_3Re'] = calculate_SB_at_R(self.entire_table.SersicIndex_r, self.entire_table.GAL_MU_E_R, 3)
            good_data_mask = good_data_mask & (self.entire_table['GAL_MU_R_at_2Re'] < 40.) & (self.entire_table['GAL_MU_R_at_2Re'] > 0.0) & (self.entire_table['GAL_MU_R_at_3Re'] < 40.) & (self.entire_table['GAL_MU_R_at_3Re'] > 0.0)
        else:
            self.entire_table['GAL_MU_R_at_2Re'] = np.nan
            self.entire_table['GAL_MU_R_at_3Re'] = np.nan


       
        # This is our entire table, which we'll now select our samples from
        self.entire_table = self.entire_table.loc[good_data_mask]

        # Check how many rows our 'entire table' has
        assert self.entire_table.shape[0]>0, "Looks like our table doesn't have any rows after getting rid of bad values of M and R_e"
        if not self.quiet:
            print(f"\nAfter getting rid of unphysical values, the master table has {self.entire_table.shape[0]} rows")



        # Get our various limits and cuts limits
        # Mag_mask is a boolean True/False which selects all galaxies brighter than this cut
        # Mstar_cut_pz is a boolean True/False which selects galaxies above pz(z) curve
        # Mstar_cut_qz is a boolean True/False which selects galaxies above qz(z) curve
        # Z_mask is a boolean True/False which selects galaxies less than our maximum z
        # Min_Re cut is a boolean True/False which selects galaxies larger than the minimum Re

        if self.BoundaryType == 'SAMI':
            self.maglimitcut = 19.8
            maglimit = f'r<{self.maglimitcut}'
            self.mag_mask = self.entire_table.GAL_MAG_R < self.maglimitcut

        elif self.BoundaryType == 'Taipan':
            self.maglimitcut = 17.0
            maglimit = f'i<{self.maglimitcut}'
            self.mag_mask = self.entire_table.GAL_MAG_I < self.maglimitcut

        elif self.BoundaryType == 'General':
            self.maglimitcut = 19.8
            maglimit = f'r<{self.maglimitcut}'
            self.mag_mask = self.entire_table.GAL_MAG_R < self.maglimitcut
            self.sf_config = self._load_selection_function_from_file(self.BoundaryFile)

        # And now select a strip in Dec, if we want to
        if np.all(np.isfinite([Dec_Min, Dec_Max])):
            self.Dec_Min = Dec_Min
            self.Dec_Max = Dec_Max
            self.Dec_mask = (self.entire_table.DEC > self.Dec_Min) & (self.entire_table.DEC < self.Dec_Max)
            if not self.quiet:
                print(f"Selecting a strip in DEC between {self.Dec_Min} and {self.Dec_Max}")
                print(f"\tIn this stripe, we have {self.entire_table.loc[self.Dec_mask].shape[0]} rows\n")

        print(f"\nMagnitude Limit: {maglimit}")

        # These are curves/lines in the M*/Z plane which tell us which galaxies to select
        # the Taipain selection is a smooth curve, SAMI is a series of steps
        qz, pz = self.get_qz_pz(self.BoundaryType)
        self.Mstar_cut_pz = self.entire_table.Mstar > pz(self.entire_table.z)
        self.Mstar_cut_qz = self.entire_table.Mstar > qz(self.entire_table.z)
        self.zcut = self.entire_table.z < self.zlimit
        self.Re_cut = self.entire_table.Re > self.minRe

        # Now we select down our samples
        # We'll have one greater than the qz(z) curve, brighter than the mag limit and z<0.1. This is self.possible_to_observe and will be plotted in RED
        # One greater than the pz(z) curve, brighter than the mag limit and less than 0.1. This is self.selection_function and will be plotted in BLUE
        # And one the same as above but sparsely sampled. This is self.selection_function_sparsely_sampled and is plotted in MAGENTA

        if np.all(np.isfinite([Dec_Min, Dec_Max])):
            self.possible_to_observe = self.entire_table.loc[self.Mstar_cut_qz & self.mag_mask & self.zcut & self.Re_cut & self.Dec_mask]
            self.selection_function = self.entire_table.loc[self.Mstar_cut_pz & self.mag_mask & self.zcut & self.Re_cut& self.Dec_mask]
        else:
            self.possible_to_observe = self.entire_table.loc[self.Mstar_cut_qz & self.mag_mask & self.zcut & self.Re_cut]
            self.selection_function = self.entire_table.loc[self.Mstar_cut_pz & self.mag_mask & self.zcut & self.Re_cut]

        nbins = int(np.ceil((self.MstarMax - self.MstarMin) * 5.))
        self.mass_bins = np.linspace(self.MstarMin, self.MstarMax, nbins + 1)
        MStarSparseCutBin = int(np.ceil((self.MSparseCut1 - self.MstarMin) * 5))

        self.selection_function_sparsely_sampled = sparse_select_flat_in_Mass(self.selection_function, self.mass_bins, MStarSparseCutBin)


    @staticmethod
    def qz_function():

        SAMIsel2_points_x = np.array([0, 0.06, 0.10, 0.15, 0.2, 0.3])
        SAMIsel2_points_y = np.array([7.0, 9.7, 9.7, 10.0, 9.7, 7.0])
        # SAMIsel2_points_x = np.array([0,0.15])
        # SAMIsel2_points_y = np.array([9.0,10.0])
        qz = np.poly1d(np.polyfit(SAMIsel2_points_x, SAMIsel2_points_y, 2))

        return qz

    @staticmethod
    def Taipan_Selection(x):
        """
        Taipan Selection cut. Lines 760-765
        """

        # Now do a selection that follows roughly the apparent stellar mass
        # completeness in the xVSMstar_SAMI_HI_iLT17.pdf TAIPAN selection -
        # which is GAMA with i<17 cut (maglimitcut value).
        SAMIsel_points_x = np.array([0.004, 0.02, 0.03, 0.04, 0.05, 0.06, 0.08,
                                     0.09, 0.1, 0.11, 0.12, 0.14, 0.20, 0.25])
        SAMIsel_points_y = np.array([7.4, 8.9, 9.6, 9.7, 10.1, 10.1, 10.2, 10.2,
                                     10.2, 10.2, 10.2, 10.3, 10.6, 10.8])

        pz = np.poly1d(np.polyfit(SAMIsel_points_x, SAMIsel_points_y, 2))

        return pz(x)

    @staticmethod
    def _load_selection_function_from_file(filename):

        with open(f"{filename}", 'r') as ymlfile:
            selection_function = yaml.safe_load(ymlfile)

        return selection_function

    def general_selection(self, z):

        if type(z) == float:
            if (z < self.sf_config['lower_lim']):
                y = self.sf_config['lower_gradient'] * z + self.sf_config['lower_constant_offset']
            if (z >= self.sf_config['lower_lim']) & (z < self.sf_config['limit_2']):
                y = self.sf_config['mass_step_2']
            if (z >= self.sf_config['limit_2']) & (z < self.sf_config['limit_3']):
                y = self.sf_config['mass_step_3']
            if (z >= self.sf_config['limit_3']) & (z < self.sf_config['limit_4']):
                y = self.sf_config['mass_step_4']
            if (z >= self.sf_config['limit_4']) & (z < self.sf_config['limit_5']):
                y = self.sf_config['mass_step_5']
            if (z >= self.sf_config['limit_5']):
                y = self.sf_config['final_mass_step']

        else:
            y = np.zeros_like(z)
            y[z < self.sf_config['lower_lim']] = self.sf_config['lower_gradient'] * z[z < self.sf_config['lower_lim']] + self.sf_config['lower_constant_offset']
            y[(z >= self.sf_config['lower_lim']) & (z < self.sf_config['limit_2'])] =self.sf_config['mass_step_2']
            y[(z >= self.sf_config['limit_2']) & (z < self.sf_config['limit_3'])] =self.sf_config['mass_step_3']
            y[(z >= self.sf_config['limit_3']) & (z < self.sf_config['limit_4'])] =self.sf_config['mass_step_4']
            y[(z >= self.sf_config['limit_4']) & (z < self.sf_config['limit_5'])] = self.sf_config['mass_step_5']
            y[(z >= self.sf_config['limit_5'])] = self.sf_config['final_mass_step']

        return y

    @staticmethod
    def sami_stepfunc(a):
        """
        Get the SAMI steps. Vectorised version of Lines 786-823.
        """

        if type(a) == float:
            if (a < 0.02):
                y = 62.5 * a + 6.95
            if (a >= 0.02) & (a < 0.03):
                y = 8.2
            if (a >= 0.03) & (a < 0.045):
                y = 8.6
            if (a >= 0.045) & (a < 0.06):
                y = 9.4
            if (a >= 0.06) & (a < 0.095):
                y = 10.3
            if (a >= 0.095):
                y = 10.9

        else:
            y = np.zeros_like(a)
            y[a < 0.02] = 62.5 * a[a < 0.02] + 6.95
            y[(a >= 0.02) & (a < 0.03)] = 8.2
            y[(a >= 0.03) & (a < 0.045)] = 8.6
            y[(a >= 0.045) & (a < 0.06)] = 9.4
            y[(a >= 0.06) & (a < 0.095)] = 10.3
            y[(a >= 0.095)] = 10.9

        return y

    def get_qz_pz(self, BoundaryType):

        qz = self.qz_function()
        if BoundaryType == "Taipan":
            return qz, self.Taipan_Selection
        elif BoundaryType == "SAMI":
            return qz, self.sami_stepfunc
        elif BoundaryType == 'General':
            return qz, self.general_selection


        # TODO: Add in sparse selection method 2

    def plot_parent_sample(self, fig, ax):
        """
        Make a plot of the parent sample in M*-z coordinates
        """

        # Linewidths and sizes of each point
        lw = 0
        s = 1

        # Check if we have surface brightness info
        if not np.all(np.isnan(self.entire_table.GAL_MU_E_R)):

            fig, ax = self._Mstar_z_plot(self.entire_table, fig, ax, 
                colour_by=self.entire_table.GAL_MU_E_R, cbar_label="Surface Brightness", vmin=20, vmax=24, lw=lw, s=s)

        else:
            fig, ax = self._Mstar_z_plot(self.entire_table, fig, ax, lw=lw, s=s)

        # Outline each point in black if we think it has HI. 
        if self.HI_info:
            HI_mask = (self.entire_table.WALLABYflag > 0) | (self.entire_table.Dingoflag > 0)
            ax.scatter(self.entire_table.z.loc[HI_mask], self.entire_table.Mstar.loc[HI_mask], lw=0.2, facecolor='None', s=s, edgecolors='k', alpha=0.5)
            ax.text(0.10, 7.5, "black=HI", color='black', fontsize=8)

        ax.text(0.005, 12.5, "Parent sample", color='green', fontsize=8)

        return fig, ax

    def plot_Mstar_z_gi_colour(self, fig, ax):
        """
        Make a plot of M* vs z and colour by G-I colour
        """

        if not np.all(np.isnan(self.entire_table.GAL_MAG_I)):
            fig, ax = self._Mstar_z_plot(self.entire_table, fig, ax, colour_by=self.entire_table['g_m_i'], cbar_label='g - i colour', vmin=0, vmax=2)
        else:
            fig, ax = self._Mstar_z_plot(self.entire_table, fig, ax)
            ax.text(0.005, 12.5, "No I-band magnitude", color='green', fontsize=8)

        return fig, ax

    def _Mstar_z_plot(self, table, fig, ax, colour_by=None, cbar_label=None, lw=0, s=1, **kwargs):

        # FIXME Add in H1 
        """
        Plot the stellar mass against redshift for the catalogue
        """
        xp = np.linspace(0, 0.15, 1000)
        # Check if we have surfave brightnesses and colour by them if we do
        if colour_by is None:
            ax.scatter(
                x=table.z, 
                y=table.Mstar,
                marker='.', lw=lw, s=s, c='grey')
            ax.scatter(
                x=table.z[self.mag_mask], 
                y=table.Mstar[self.mag_mask], 
                marker='o', lw=lw, s=s, c='green')

        else:
            ax.scatter(
                x=table.z, 
                y=table.Mstar, 
                c=colour_by, 
                cmap=self.cm, marker='.', lw=lw, s=s, **kwargs)

            sc = ax.scatter(
                x=table.z[self.mag_mask], 
                y=table.Mstar[self.mag_mask], 
                c=colour_by[self.mag_mask], 
                cmap=self.cm, marker='o', lw=lw, s=s, **kwargs)
            cbar = fig.colorbar(sc, ax=ax)
            cbar.set_label(cbar_label)

        qz, pz = self.get_qz_pz(self.BoundaryType)

        ax.plot(xp, pz(xp), '-', color='blue')
        # plt.plot(xp, pz(xp), '-', color='orange')
        ax.plot(xp, qz(xp), '-', color='red')
        ax.set_ylim(7, 13)
        ax.set_xlim(0, 0.15)
        ax.set_xlabel('z')
        ax.set_ylabel('Mstar')

        return fig, ax

    def plot_z_Re(self, fig, ax):

        sc = ax.scatter(
            x=self.entire_table.z[self.mag_mask], 
            y=self.entire_table.Re[self.mag_mask], 
            c=(self.entire_table.g_m_i)[self.mag_mask], 
            vmin=0, vmax=2.0, cmap=self.cm, marker='o', lw=0, s=1)

        cbar = fig.colorbar(sc, ax=ax)
        cbar.set_label('g-i colour')

        # plt.plot(xp, pz(xp), '-', color='blue')
        # plt.plot(xp, pz(xp), '-', color='orange')
        # plt.plot(xp, qz(xp), '-', color='red')
        ax.set_ylim(0, 15)
        ax.set_xlim(0, 0.15)
        ax.set_xlabel('z')
        ax.set_ylabel('Re')

        return fig, ax

    def plot_selected_sample_MZ(self, fig, ax):
        """
        Make another M*-Z plot, but this time mark on each sample. This is going to be a new function which doesn't reusue _plot_M_z- I can't see an easy way to do what I want with that function without changing everything!
        """

        ax.scatter(self.entire_table.z, self.entire_table.Mstar, color='grey', marker='.')
        ax.scatter(self.possible_to_observe.z, self.possible_to_observe.Mstar, color='red', marker='.', lw=0, s=2)
        ax.scatter(self.selection_function.z, self.selection_function.Mstar, color='blue', marker='o', lw=0, s=7)
        ax.scatter(self.selection_function_sparsely_sampled.z, self.selection_function_sparsely_sampled.Mstar, color='magenta', marker='o', lw=0, s=7)

        if self.HI_info:
            HI_mask = (self.selection_function_sparsely_sampled.WALLABYflag > 0) | (self.selection_function_sparsely_sampled.Dingoflag > 0)
            ax.scatter(self.selection_function_sparsely_sampled.z.loc[HI_mask], self.selection_function_sparsely_sampled.Mstar.loc[HI_mask], color='black', marker='o', s=1)
            ax.text(0.45, 0.1, "Black: Magenta sample\nwith HI", color='black', transform=ax.transAxes, fontsize=8, bbox=dict(facecolor='white', edgecolor='k', alpha=0.8, boxstyle='round,pad=0.1'))

        xp = np.linspace(0, 0.15, 1000)
        qz, pz = self.get_qz_pz(self.BoundaryType)

        ax.plot(xp, pz(xp), '-', color='blue')
        # plt.plot(xp, pz(xp), '-', color='orange')
        ax.plot(xp, qz(xp), '-', color='red')

        ax.set_ylim(7, 13)
        ax.set_xlim(0, 0.15)
        ax.set_xlabel('z')
        ax.set_ylabel('Mstar')
        ax.text(0.05, 0.9, "Selected samples", color='green', transform=ax.transAxes, fontsize=8)
        
        return fig, ax

    def plot_selected_sample_MZ_colour_Re(self, fig, ax):

        cmap_kwargs = dict(vmin=0, vmax=5, cmap=self.cm)
        sc = ax.scatter(self.possible_to_observe.z, self.possible_to_observe.Mstar, c=self.possible_to_observe.Re, marker='.', linewidths=0, s=0.5, **cmap_kwargs)
        sc = ax.scatter(self.selection_function.z, self.selection_function.Mstar, c=self.selection_function.Re, marker='.', linewidths=0, s=1, **cmap_kwargs)
        sc = ax.scatter(self.selection_function_sparsely_sampled.z, self.selection_function_sparsely_sampled.Mstar, c=self.selection_function_sparsely_sampled.Re, marker='.', linewidths=0, s=2, **cmap_kwargs)

        cbar = fig.colorbar(sc, ax=ax)

        Frac_Re_Cut = 2  # Hardcoded number? Line 1171 in Original code
        cbar.set_label(f'Re (% Re>{Frac_Re_Cut} in black boxes')

        # Add the lines
        qz, pz = self.get_qz_pz(self.BoundaryType)
        xp = np.linspace(0, 0.15, 1000)

        # Not sure why +0.05 here?
        ax.plot(xp, pz(xp) + 0.05, '-', color='cyan')

        ax.plot(xp[xp <= 0.1], pz(xp)[xp <= 0.1], '-', color='blue')
        ax.plot(xp[xp <= 0.1], qz(xp)[xp <= 0.1], '-', color='red')
        ax.plot([0.1, 0.1], [pz(0.1), 12], '-', color='blue')
        ax.plot([0.1005, 0.1005], [qz(0.1), 12], '-', color='red')
        ax.plot([self.zlimit, self.zlimit], [pz(self.zlimit), 12], '-', color='magenta')
        ax.plot([0.03, 0.045], [8.6, 8.6], '-', color='black')  # plotting SAMI filler lower limits
        ax.plot([0.045, 0.06], [9.4, 9.4], '-', color='black')  # plotting SAMI filler lower limits
        ax.plot([0.06, 0.095], [10.3, 10.3], '-', color='black')  # plotting SAMI filler lower limits
        ax.plot([0.095, 0.115], [10.9, 10.9], '-', color='black')  # plotting SAMI filler lower limits

        # Add in the lines and fractions of galaxies in each box
        fig, ax = add_boxes_to_Mstar_z(fig, ax, self.selection_function)

        ax.set_ylim(7, 13)
        ax.set_xlim(0, 0.15)
        ax.set_xlabel('z')
        ax.set_ylabel('Mstar')

        # Add in the source densities
        if self.total_area > 0:
            density_in_selection = np.round(self.selection_function.shape[0] * np.pi / self.total_area, decimals=0)
            density_above_SB_limit = np.round(self.possible_to_observe.shape[0] * np.pi / self.total_area, decimals=0)
            density_in_sparse_sampled_selection = np.round(self.selection_function_sparsely_sampled.shape[0] * np.pi / self.total_area, decimals=0)
        else:
            density_in_selection = 'N/A'
            density_above_SB_limit = 'N/A'
            density_in_sparse_sampled_selection = 'N/A'

        ax.text(0.11, 7.5, density_above_SB_limit, color='red')
        ax.text(0.05, 7.5, density_in_selection, color='blue')
        ax.text(0.08, 7.5, density_in_sparse_sampled_selection, color='magenta')
        ax.text(0.04, 8.0, "Gal/2dF(3.14sq.deg)=", color='black')
        return fig, ax

    def _plot_hist(self, column_name, bins, fig, ax):
        """
        Plot a histogram of a column.
        If given, mask should be a list of three masks for each table. 

        """

        red = self.possible_to_observe[f'{column_name}']
        blue = self.selection_function[f'{column_name}']
        magenta = self.selection_function_sparsely_sampled[f'{column_name}']

        ax.hist(red, bins=bins, facecolor='r')    
        ax.hist(blue, bins=bins, facecolor='b')
        ax.hist(magenta, bins=bins, facecolor='magenta')
        ax.set_ylabel("No. galaxies")

        return fig, ax

    def plot_mass_hist(self, fig, ax):

        fig, ax = self._plot_hist('Mstar', self.mass_bins, fig, ax)
        ax.set_xlabel("Stellar Mass")
        return fig, ax

    def plot_z_hist(self, fig, ax):

        bins = np.linspace(0.0, 0.12, 31)
        fig, ax = self._plot_hist('z', bins, fig, ax)
        ax.set_xlabel("Redshift")
        return fig, ax

    def plot_Re_hist(self, fig, ax):

        Remin = 0
        Remax = 15
        bins = np.linspace(Remin, Remax, int(np.ceil((Remax - Remin) * 4.)))

        fig, ax = self._plot_hist('Re', bins, fig, ax)

        ax.text(0.7, 0.92, "median Re", color='black', fontsize=8, transform=ax.transAxes)  # was 13, 600
        ax.text(0.75, 0.85, np.round(np.nanmedian(self.possible_to_observe.Re), decimals=2), color='red', transform=ax.transAxes)
        ax.text(0.75, 0.8, np.round(np.nanmedian(self.selection_function.Re), decimals=2), color='blue', transform=ax.transAxes)
        ax.text(0.75, 0.75, np.round(np.nanmedian(self.selection_function_sparsely_sampled.Re), decimals=2), color='magenta', transform=ax.transAxes)

        ax.set_xlabel("R_e (arcsec)")
        return fig, ax

    def plot_gmi_hist(self, fig, ax):

        bins = np.linspace(0, 5, 21)
        fig, ax = self._plot_hist('g_m_i', bins, fig, ax)

        ax.set_xlabel("g-i colour")
        ax.set_xlim(0, 3)
        return fig, ax

    def plot_ellipticity_hist(self, fig, ax):

        bins = np.linspace(0, 1, 21)
        fig, ax = self._plot_hist('Ellipticity_r', bins, fig, ax)

        ax.set_xlabel("Ellipticity")
        return fig, ax

    def plot_SB_Re(self, fig, ax):

        # plt.plot(HECTORdata_Retemp,HECTORdata_SB, '.', color='magenta')
        ax.plot(self.selection_function.Re, self.selection_function.GAL_MU_E_R, 'o', color='blue')
        ax.plot(self.selection_function_sparsely_sampled.Re, self.selection_function_sparsely_sampled.GAL_MU_E_R, '.', color='magenta')
        ax.set_xlim(0, 15)
        ax.set_ylim(10, 30)

        ax.set_xlabel('Re (arcsec)')
        ax.set_ylabel('Surface Brightness\n (mag/arcsec^2)')

        return fig, ax

    def plot_IFU_needed_for_2Re_hist(self, fig, ax):

        fig, ax = self._plot_hist('IFU_diam_2Re', bins=100, fig=fig, ax=ax)
        ax.set_xlim(0, 60.0)
        ax.set_xlabel("IFU diameter required for 2Re (arcsec)")
        return fig, ax

    def _get_fraction_of_sample_which_fits_in_IFU_diameter(self):
        """
        Bin up the sample using np.histogram in terms of the diameter of a 2Re circle for each galaxy. We then work out the fraction of the sample which is less than each diameter (by dividing by N_total). This is a bit like a KDE of table.IFU_diam_2Re, but not quite! KDEs integrate to 1 whereas these values _sum_ to 1
        Order is [red, blue, magenta]

        Returns: A list of fractions, A list of bin edges.
        """

        N_pto = self.possible_to_observe.IFU_diam_2Re.size
        pto_binned_IFU_diam, bin_edges_pto = np.histogram(self.possible_to_observe.IFU_diam_2Re, density=False, bins=200)

        N_sf = self.selection_function.IFU_diam_2Re.size
        sf_binned_IFU_diam, bin_edges_sf = np.histogram(self.selection_function.IFU_diam_2Re, density=False, bins=200)

        N_sfss = self.selection_function_sparsely_sampled.IFU_diam_2Re.size
        sfss_binned_IFU_diam, bin_edges_sfss = np.histogram(self.selection_function_sparsely_sampled.IFU_diam_2Re, density=False, bins=200)

        fractions = [pto_binned_IFU_diam / N_pto, sf_binned_IFU_diam / N_sf, sfss_binned_IFU_diam / N_sfss]
        bins = [bin_edges_pto, bin_edges_sf, bin_edges_sfss]

        return fractions, bins

    def plot_IFU_needed_for_2Re_kde(self, fig, ax):

        fractions, bins = self._get_fraction_of_sample_which_fits_in_IFU_diameter()

        bin_edges_pto, bin_edges_sf, bin_edges_sfss = bins
        fractions_pto, fractions_sf, fractions_sfss = fractions

        ax.plot(bin_edges_sfss[:-1], fractions_sfss, color='magenta')
        ax.plot(bin_edges_sf[:-1], fractions_sf, color='blue')
        ax.plot(bin_edges_pto[:-1], fractions_pto, color='red')

        ax.set_xlabel("Bundle diameter to reach 2Re (arcsec)")
        ax.set_ylabel("Cumulative frac. sample\n for 2Re every galaxy")

        ax.set_xlabel("IFU diameter required for 2Re (arcsec)")
        ax.set_ylabel("Fraction of Sample")
        ax.set_xlim(0, 100)
        return fig, ax

    def plot_cumulative_fraction_imaged_vs_bundle_diameter(self, fig, ax):

        ax.set_xlim(0, 100)
        ax.text(16, 0.15, "127", color='black')
        ax.text(10, 0.25, "num. 105um fibres/bundle", color='black')
        ax.text(36, 0.15, "469", color='black')
        ax.text(55, 0.15, "1027", color='black')
        ax.text(75, 0.15, "1801", color='black')

        fractions, bins = self._get_fraction_of_sample_which_fits_in_IFU_diameter()

        bin_edges_pto, bin_edges_sf, bin_edges_sfss = bins
        fractions_pto, fractions_sf, fractions_sfss = fractions

        # cumulative_IFU_diam_pto = np.cumsum(fractions_pto)
        cumulative_IFU_diam_sf = np.cumsum(fractions_sf)
        cumulative_IFU_diam_sfss = np.cumsum(fractions_sfss)

        ax.plot(bin_edges_sf[:-1], cumulative_IFU_diam_sf, color='blue')
        ax.plot(bin_edges_sfss[:-1], cumulative_IFU_diam_sfss, color='magenta')
        ax.set_xlabel("Bundle diameter to reach 2Re (arcsec)")
        ax.set_ylabel("Cumulative frac. sample\n for 2Re every galaxy")

        return fig, ax

    def cumulative_fraction_sample_imaged_to_2Re(self, fig, ax):
        """
        Lines 1435-1486
        """

        fractions, bins = self._get_fraction_of_sample_which_fits_in_IFU_diameter()

        bin_edges_pto, bin_edges_sf, bin_edges_sfss = bins
        fractions_pto, fractions_sf, fractions_sfss = fractions

        DIA = np.array([1.676190476, 5.028571429, 8.380952381, 11.73333333, 15.08571429, 18.43809524, 21.79047619, 25.14285714, 28.4952381, 31.84761905, 35.2, 38.55238095, 41.9047619, 45.25714286, 48.60952381, 51.96190476, 55.31428571, 58.66666667, 62.01904762, 65.37142857, 68.72380952, 72.07619048, 75.42857143, 78.78095238, 82.13333333, 85.48571429, 88.83809524, 92.19047619, 95.54285714, 98.8952381, 102.247619, 105.6])  # assumes 105um cores with 5um cladding.
        # coresize = 105
        CORES = np.array([1, 7, 19, 37, 61, 91, 127, 169, 217, 271, 331, 397, 469, 547, 631, 721, 817, 919, 1027, 1141, 1261, 1387, 1519, 1657, 1801, 1951, 2107, 2269, 2437, 2611, 2791, 2977])

        # Now want cumulative fraction of bundles at each diameter and plot
        # Set maximum bundle size that is likely as 469=core, or the 13th element of the CORES and DIA arrays, which is overkill, but set here (-1 because 0 to 12 array):
        MAX_bundle = 13 - 1
        coresplot = CORES[0:MAX_bundle]
        bundles_frac_magenta = []  # to make a zero length array that I will append to during the loop.
        bundles_frac_blue = []
        for i in range(0, MAX_bundle):
            # test = np.where(CORES==i) #find the index at position i, as a tuple
            # tempindex = np.int(test[0]) # make theindex an integer
            # tempDIA[i] = DIA[tempindex[i]] # find the DIA value for that index

            bundles = bisect.bisect(bin_edges_sf[:-1], DIA[i]) - 1  # find first bin in the cumulative bundle diameter for 2Re plot
            # print "bundles", bundles
            # print "binsb", binsb
            # print "DIA", DIA[i]
            bundles_zLT0pt1 = bisect.bisect(bin_edges_sfss[:-1], DIA[i]) - 1
            # Now sum the fract_bundlediam_2Re fraction of galaxies at 2Re from the 0th element to the "bundles" element (should be equivalent to the bundles-th element of cumulative array:
            bundles_frac_magenta.append(np.sum(fractions_sfss[0:(bundles + 1)]))
            # print "bundles_frac", bundles_frac
            bundles_frac_blue.append(np.sum(fractions_sf[0:(bundles_zLT0pt1 + 1)]))
            # bundlesar = np.asarray(bundles)
            # for j in bundles:
        # total_fraction = cumulative[j] #find the total fraction of galaxies with 2Re<the diameter of the bundle for all bundle sizes. eg: all 2Re bins smaller than an nth-core hexabundle, are summed, so we get the cumulative fraction of galaxies that would fit within that bundle size.
        bundles_frac_magenta = np.hstack(bundles_frac_magenta)  # This converts the list to an array.
        # print "bundles_frac_arr", bundles_frac_arr
        bundles_frac_blue = np.hstack(bundles_frac_blue)

        # Now plot the bundles size in no. of cores vs fraction of galaxies:
        ax.plot(coresplot, bundles_frac_magenta, color='magenta')
        ax.plot(coresplot, bundles_frac_blue, color='blue')
        ax.text(19, 0.15, "19", color='black', ha='center')
        ax.plot([19, 19], [0.2, 1.0], ':', color='black')
        ax.text(37, 0.15, "37", color='black', ha='center')
        ax.plot([37, 37], [0.2, 1.0], ':', color='black')
        ax.text(61, 0.15, "61", color='black', ha='center')
        ax.plot([61, 61], [0.2, 1.0], ':', color='black')
        ax.text(91, 0.15, "91", color='black', ha='center')
        ax.plot([91, 91], [0.2, 1.0], ':', color='black')
        ax.text(127, 0.15, "127", color='black', ha='center')
        ax.plot([127, 127], [0.2, 1.0], ':', color='black')
        ax.text(169, 0.15, "169", color='black', ha='center')
        ax.plot([169, 169], [0.2, 1.0], ':', color='black')
        ax.text(217, 0.15, "217", color='black', ha='center')
        ax.plot([217, 217], [0.2, 1.0], ':', color='black')
        ax.set_ylabel("cumulative frac. of galaxies")
        ax.set_xlabel("No.cores in the bundle that just fits 2Re")

        return fig, ax

    def plot_SB_hist(self, fig, ax):
        """
        Plot histograms of effective surface brightness within 1Re and surface brightness at 2 and 3Re. 
        """

        bins = 50
        # Check if surface brightness is nan, as otherwise hist complains
        if not np.all(np.isnan(self.possible_to_observe.GAL_MU_E_R)):
            # Check if we've been able to make surface brightness measurements at larger radii too
            if not np.all(np.isnan(self.possible_to_observe.GAL_MU_R_at_2Re)):

                ax.hist(self.possible_to_observe.GAL_MU_R_at_3Re, color='red', bins=bins, alpha=0.2)
                ax.hist(self.selection_function.GAL_MU_R_at_3Re, color='blue', bins=bins, alpha=0.2)
                ax.hist(self.selection_function_sparsely_sampled.GAL_MU_R_at_3Re, color='magenta', bins=bins, alpha=0.2)

                ax.hist(self.possible_to_observe.GAL_MU_R_at_2Re, color='red', bins=bins, alpha=0.5)
                ax.hist(self.selection_function.GAL_MU_R_at_2Re, color='blue', bins=bins, alpha=0.5)
                ax.hist(self.selection_function_sparsely_sampled.GAL_MU_R_at_2Re, color='magenta', bins=bins, alpha=0.5)
            ax.hist(self.possible_to_observe.GAL_MU_E_R, color='red', bins=bins)
            ax.hist(self.selection_function.GAL_MU_E_R, color='blue', bins=bins)
            ax.hist(self.selection_function_sparsely_sampled.GAL_MU_E_R, color='magenta', bins=bins)

        else:   
            # Plot some fake data otherwise matplotlib complains when I try and    
            # ax.hist(np.random.normal(loc=20, scale=3, size=100), alpha=0.0)      
            ax.set_ylim(0, 250)

        ax.set_xlim(18, 28)
        ax.set_xlabel("Surface brightness at 1Re (filled) \n and 2Re (tranparent)\n and 3Re (most transparent) (mag/arcsec^2)")
        ax.set_ylabel('No. galaxies')

        # Line at the limiting surface brightness of a SAMI galaxy to get S/N ~5 after 4 hours when when binned up 
        # by 24 cores
        # ax.axvline(25, '-', color='green')
        
        return fig, ax

    def plot_percentage_galaxy_sample_imaged_2Re(self, fig, ax):
        """
        Lines 1597 - 1707
        """

        # TotalFracto2Re=((percent100umcores[7,1]+percent100umcores[8,1]) - percentfinalbundledistribution[7,1])+ (percent100umcores[6,1]-(percentfinalbundledistribution[6,1]-(percent100umcores[7,1] - percentfinalbundledistribution[7,1]))) + (percent100umcores[5,1]-(percentfinalbundledistribution[5,1]-((percent100umcores[6,1]-(percentfinalbundledistribution[6,1]-(percent100umcores[7,1] - percentfinalbundledistribution[7,1])))))) + (percent100umcores[4,1]-(percentfinalbundledistribution[4,1]- ((percent100umcores[5,1]-(percentfinalbundledistribution[5,1]-((percent100umcores[6,1]-(percentfinalbundledistribution[6,1]-(percent100umcores[7,1] - percentfinalbundledistribution[7,1])))))))))

        percent100umcores, percent100umcores_zLT0pt1, percent75umcores, cores9, cores13, TotalFracto2Re = self._get_percent_100um_cores()

        # ax = fig3.add_subplot(2,4,8)
        ax.plot(cores9, percent100umcores[:, 1], '-o', color='magenta', label='105um fibres')
        ax.plot(cores9, percent100umcores_zLT0pt1[:, 1], '-o', color='blue', label='105um fibres')
        # ax.plot(cores13,percent75umcores[:,1], '-.', color='magenta',label='75um fibres')
        ax.text(17, 27, "19", color='black', size=8, ha='center')
        ax.plot([19, 19], [2, 20], ':', color='black')
        ax.text(37, 27, "37", color='black', size=8, ha='center')
        ax.plot([37, 37], [2, 20], ':', color='black')
        ax.text(61, 27, "61", color='black', size=8, ha='center')
        ax.plot([61, 61], [2, 20], ':', color='black')
        ax.text(91, 27, "91", color='black', size=8, ha='center')
        ax.plot([91, 91], [2, 20], ':', color='black')
        ax.text(127, 27, "127", color='black', size=8, ha='center')
        ax.plot([127, 127], [2, 20], ':', color='black')
        ax.text(169, 27, "169", color='black', size=8, ha='center')
        ax.plot([169, 169], [2, 20], ':', color='black')
        ax.text(217, 27, "217", color='black', size=8, ha='center')
        ax.plot([217, 217], [2, 20], ':', color='black')
        ax.text(50, 3, "Percentage (magenta)\nimaged to 2Re = " + str(TotalFracto2Re) + "%", color='magenta', size=8, ha='left', bbox=dict(facecolor='white', edgecolor='black', alpha=0.5, boxstyle='round,pad=0.1'))
        ax.set_xlim(0, 300)
        ax.set_ylim(0, 30)
        ax.set_xlabel('No.cores in the bundle that just fits 2Re')
        ax.set_ylabel('% galaxy sample \n imaged to 2Re')

        return fig, ax



    def _get_percent_100um_cores(self):

        bundlediam_2Re = self.selection_function_sparsely_sampled['IFU_diam_2Re']
        bundlediam_2Re_zLT0pt1 = self.selection_function['IFU_diam_2Re']

        cores9 = [1, 7, 19, 37, 61, 91, 127, 169, 170]  # 217, 218]
        cores13 = [1, 7, 19, 37, 61, 91, 127, 169, 217, 271, 331, 397, 469, 470]
        # diamfor100umcores = [1.726476, 5.17942857, 8.63238, 12.085333, 15.5382857, 18.99123809, 22.44419, 25.89714]  # , 28.4952381]  updated for 103um core and max 169 cores
        # diamfor75umcores = [1.219047619, 3.657142857, 6.095238095, 8.533333333, 10.97142857, 13.40952381, 15.84761905, 18.28571429, 20.72380952, 23.16190476, 25.6, 28.03809524, 30.47619048]
        # percentcores = [0.3, 3.5, 9.5, 13.8, 15.6, 13, 11, 9, 6]
        # plt.plot(cores, percentcores) which I quickly plotted for my instrument science talk.
        # Make a temporary array for the number of galaxies for each of these 9 diameters plus an extra one for > 28.5
        percent100umcores = np.array([[1.726476, 1], [5.17942857, 1], [8.63238, 1], [12.085333, 1], [15.5382857, 1], [18.99123809, 1], [22.44419, 1], [25.89714, 1], [28, 1]])  # Note:points at 28 are for everything >25.897
        percent100umcores_zLT0pt1 = np.array([[1.726476, 1], [5.17942857, 1], [8.63238, 1], [12.085333, 1], [15.5382857, 1], [18.99123809, 1], [22.44419, 1], [25.89714, 1], [28, 1]])  # Note:points at 30 are for everything >28.5
        percent75umcores = np.array([[1.219047619, 1], [3.657142857, 1], [6.095238095, 1], [8.533333333, 1], [10.97142857, 1], [13.40952381, 1], [15.84761905, 1], [18.28571429, 1], [20.72380952, 1], [23.16190476, 1], [25.6, 1], [28.03809524, 1], [30.47619048, 1], [35, 1]])  # Note:points at 35 are for everything >30.5
        lenpc100 = len(percent100umcores)
        lenpc75 = len(percent75umcores)

        percentfinalbundledistribution = np.array([[1.726476, 0], [5.17942857, 0], [8.63238, 0], [12.085333, 0], [15.5382857, (14. / 19.) * 100], [18.99123809, (2. / 19.) * 100], [22.44419, (1. / 19.) * 100], [25.89714, (2. / 19.) * 100], [28, 0.]])  # excluding the 2 star bundles

        for i in range(lenpc100):
            if i == 0:
                test1 = float(len(bundlediam_2Re[(bundlediam_2Re < percent100umcores[i, 0])]))
                test3 = ((test1) / (float(len(bundlediam_2Re)))) * 100
                percent100umcores[i, 1] = test3
                test1 = float(len(bundlediam_2Re_zLT0pt1[(bundlediam_2Re_zLT0pt1<percent100umcores[i,0])]))
                test3 = ((test1)/(float(len(bundlediam_2Re_zLT0pt1))))*100
                percent100umcores_zLT0pt1[i,1] = test3
            if i>0:
                test1 = float(len(bundlediam_2Re[(bundlediam_2Re<percent100umcores[i,0])]))
                test2 = float(len(bundlediam_2Re[(bundlediam_2Re<percent100umcores[i-1,0])]))
                test3 = ((test1 - test2)/(float(len(bundlediam_2Re))))*100
                percent100umcores[i,1] = test3
                test1 = float(len(bundlediam_2Re_zLT0pt1[(bundlediam_2Re_zLT0pt1<percent100umcores[i,0])]))
                test2 = float(len(bundlediam_2Re_zLT0pt1[(bundlediam_2Re_zLT0pt1<percent100umcores[i-1,0])]))
                test3 = ((test1 - test2)/(float(len(bundlediam_2Re_zLT0pt1))))*100
                percent100umcores_zLT0pt1[i,1] = test3
            if i==(lenpc100-1):
                test1 = float(len(bundlediam_2Re[(bundlediam_2Re>percent100umcores[i-1,0])]))
                test3 = ((test1)/(float(len(bundlediam_2Re))))*100
                percent100umcores[i,1] = test3
                test1 = float(len(bundlediam_2Re_zLT0pt1[(bundlediam_2Re_zLT0pt1>percent100umcores[i-1,0])]))
                test3 = ((test1)/(float(len(bundlediam_2Re_zLT0pt1))))*100
                percent100umcores_zLT0pt1[i,1] = test3
        #print percent100umcores
        for i in range(lenpc75):
            if i==0:
                test1= float(len(bundlediam_2Re[(bundlediam_2Re<percent75umcores[i,0])]))
                test3 = ((test1)/(float(len(bundlediam_2Re))))*100
                percent75umcores[i,1] = test3
            if i>0:
                test1 = float(len(bundlediam_2Re[(bundlediam_2Re<percent75umcores[i,0])]))
                test2 = float(len(bundlediam_2Re[(bundlediam_2Re<percent75umcores[i-1,0])]))
                test3 = ((test1 - test2)/(float(len(bundlediam_2Re))))*100
                percent75umcores[i,1] = test3
            if i == (lenpc75-1):
                test1 = float(len(bundlediam_2Re[(bundlediam_2Re>percent75umcores[i-1,0])]))
                test3 = ((test1)/(float(len(bundlediam_2Re))))*100
                percent75umcores[i,1] = test3
        #print percent75umcores

        # Now calculate the fraction of all galaxies that are imaged to 2Re if the largest is put in the largest bundle etc, down to the smallest - i.e. not allowing for tiling variations.
        #Total
        if  percent100umcores[8,1]> percentfinalbundledistribution[7,1]:
            numfit169 = 0
            numleft169 = percent100umcores[8,1] - percentfinalbundledistribution[7,1] + percent100umcores[7,1]
        else:
            numfit169 = percentfinalbundledistribution[7,1]-percent100umcores[8,1]
            numleft169 = percent100umcores[7,1] - numfit169 #number in the 169 bin that did not fit
        if  numleft169 > percentfinalbundledistribution[6,1]:
            numfit127 = 0
            numleft127 = numleft169 - percentfinalbundledistribution[6,1] + percent100umcores[6,1]
        else:
            numfit127 = percentfinalbundledistribution[6,1]- numleft169
            numleft127 = numleft169 - numfit127 + percent100umcores[6,1]
        if  numleft127 > percentfinalbundledistribution[5,1]:
            numfit91 = 0
            numleft91 = numleft127 - percentfinalbundledistribution[5,1] + percent100umcores[5,1]
        else:
            numfit91 = percentfinalbundledistribution[5,1]- numleft127
            numleft91 = numleft127 - numfit91 + percent100umcores[5,1]
        if  numleft91 > percentfinalbundledistribution[4,1]:
            numfit61 = 0
            numleft61 = numleft91 - percentfinalbundledistribution[4,1] + percent100umcores[4,1]
        else:
            numfit61 = percentfinalbundledistribution[4,1]- numleft91
            numleft61 = numleft91 - numfit61 + percent100umcores[4,1]
        TotalFracto2Re = np.int(numfit169 + numfit127 + numfit91 + numfit61)

        return percent100umcores, percent100umcores_zLT0pt1, percent75umcores, cores9, cores13, TotalFracto2Re



    def cumulative_percentage_galaxies_imaged_2Re(self, fig, ax):
        """
        Lines 1716- 1798
        """

        percent100umcores, percent100umcores_zLT0pt1, percent75umcores, cores9, cores13, TotalFracto2Re = self._get_percent_100um_cores()

        percent100umcores[4,1] = percent100umcores[0,1] + percent100umcores[1,1] + percent100umcores[2,1] + percent100umcores[3,1] + percent100umcores[4,1]
        percent100umcores[0,1] = 0
        percent100umcores[1,1] = 0
        percent100umcores[2,1] = 0
        percent100umcores[3,1] = 0
        #print "percent100umcores",percent100umcores
        percent75umcores[6,1] = percent75umcores[0,1] + percent75umcores[1,1] + percent75umcores[2,1] + percent75umcores[3,1] + percent75umcores[4,1] + percent75umcores[5,1] + percent75umcores[6,1]
        percent75umcores[0,1] = 0
        percent75umcores[1,1] = 0
        percent75umcores[2,1] = 0
        percent75umcores[3,1] = 0
        percent75umcores[4,1] = 0
        percent75umcores[5,1] = 0
        #***


        #Now do cumulative plot vs cumulative number of cores:
        cum_pc100umcores = np.cumsum(percent100umcores[:,1])
        #print "cum_pc100umcores", cum_pc100umcores
        cum_pc75umcores = np.cumsum(percent75umcores[:,1])
        totalbundles = np.array([100,90,80,70,60,50,20]) #total number of hexabundles
        cores9_sumfibrenumbers_100 = cores9 * (percent100umcores[:,1] /100) * totalbundles[0]
        cumcores9_sumfibrenumbers_100 = np.cumsum(cores9_sumfibrenumbers_100)
        cores9_sumfibrenumbers_90 = cores9 * (percent100umcores[:,1] /100) * totalbundles[1]
        cumcores9_sumfibrenumbers_90 = np.cumsum(cores9_sumfibrenumbers_90)
        cores9_sumfibrenumbers_80 = cores9 * (percent100umcores[:,1] /100) * totalbundles[2]
        cumcores9_sumfibrenumbers_80 = np.cumsum(cores9_sumfibrenumbers_80)
        cores9_sumfibrenumbers_70 = cores9 * (percent100umcores[:,1] /100) * totalbundles[3]
        cumcores9_sumfibrenumbers_70 = np.cumsum(cores9_sumfibrenumbers_70)
        cores9_sumfibrenumbers_60 = cores9 * (percent100umcores[:,1] /100) * totalbundles[4]
        cumcores9_sumfibrenumbers_60 = np.cumsum(cores9_sumfibrenumbers_60)
        cores9_sumfibrenumbers_50 = cores9 * (percent100umcores[:,1] /100) * totalbundles[5]
        cumcores9_sumfibrenumbers_50 = np.cumsum(cores9_sumfibrenumbers_50)
        cores9_sumfibrenumbers_20 = cores9 * (percent100umcores[:,1] /100) * totalbundles[6]
        cumcores9_sumfibrenumbers_20 = np.cumsum(cores9_sumfibrenumbers_20)
        percentintervals=np.array([0,5,10,15,20,25,30,35,40,45,50,55,60,65,70,75,80,85,90,95,100])
        percentintervals_100um_50 = np.interp(percentintervals,cum_pc100umcores, cumcores9_sumfibrenumbers_50)#print the cumulative number of cores at 90% to work out total angular area. Note had to reverse x and y
        percentintervals_100um_100 = np.interp(percentintervals,cum_pc100umcores, cumcores9_sumfibrenumbers_100)
        # print "Number of 105um cores in 50 hexas when", percentintervals[18],"% of galaxies imaged to 2Re is",percentintervals_100um_50[18],"which is a total fibre area of", percentintervals_100um_50[18] * pi * (0.8**2.), "arcsec^2"
        # print "Number of 105um cores in 100 hexas when", percentintervals[18],"% of galaxies imaged to 2Re is",percentintervals_100um_100[18],"which is a total fibre area of", percentintervals_100um_100[18] * pi * (0.8**2.), "arcsec^2"


        cores13_sumfibrenumbers_100 = cores13 * (percent75umcores[:,1] /100) * totalbundles[0]
        cumcores13_sumfibrenumbers_100 = np.cumsum(cores13_sumfibrenumbers_100)
        cores13_sumfibrenumbers_90 = cores13 * (percent75umcores[:,1] /100) * totalbundles[1]
        cumcores13_sumfibrenumbers_90 = np.cumsum(cores13_sumfibrenumbers_90)
        cores13_sumfibrenumbers_80 = cores13 * (percent75umcores[:,1] /100) * totalbundles[2]
        cumcores13_sumfibrenumbers_80 = np.cumsum(cores13_sumfibrenumbers_80)
        cores13_sumfibrenumbers_70 = cores13 * (percent75umcores[:,1] /100) * totalbundles[3]
        cumcores13_sumfibrenumbers_70 = np.cumsum(cores13_sumfibrenumbers_70)
        cores13_sumfibrenumbers_60 = cores13 * (percent75umcores[:,1] /100) * totalbundles[4]
        cumcores13_sumfibrenumbers_60 = np.cumsum(cores13_sumfibrenumbers_60)
        cores13_sumfibrenumbers_50 = cores13 * (percent75umcores[:,1] /100) * totalbundles[5]
        cumcores13_sumfibrenumbers_50 = np.cumsum(cores13_sumfibrenumbers_50)
        cores13_sumfibrenumbers_20 = cores13 * (percent75umcores[:,1] /100) * totalbundles[5]
        cumcores13_sumfibrenumbers_20 = np.cumsum(cores13_sumfibrenumbers_20)
        percentintervals=np.array([0,5,10,15,20,25,30,35,40,45,50,55,60,65,70,75,80,85,90,95,100])
        percentintervals_75um_50 = np.interp(percentintervals,cum_pc75umcores, cumcores13_sumfibrenumbers_50)#print the cumulative number of cores at 90% to work out total angular area. Note had to reverse x and y
        percentintervals_75um_100 = np.interp(percentintervals,cum_pc75umcores, cumcores13_sumfibrenumbers_100)
        # print "Number of 75um cores in 50 hexas when", percentintervals[18],"% of galaxies imaged to 2Re is",percentintervals_75um_50[18],"which is a total fibre area of", percentintervals_75um_50[18] * pi * ((0.75*0.8)**2.), "arcsec^2"
        # print "Number of 75um cores in 100 hexas when", percentintervals[18],"% of galaxies imaged to 2Re is",percentintervals_75um_100[18],"which is a total fibre area of", percentintervals_75um_100[18] * pi * ((0.75*0.8)**2.), "arcsec^2"

        #plt.plot(cumcores9_sumfibrenumbers_100,cum_pc100umcores, '-o', color='green',label='100')
        #plt.plot(cumcores9_sumfibrenumbers_90,cum_pc100umcores, '-o', color='cyan',label='90')
        #plt.plot(cumcores9_sumfibrenumbers_80,cum_pc100umcores, '-o', color='blue',label='80')
        #plt.plot(cumcores9_sumfibrenumbers_70,cum_pc100umcores, '-o', color='red',label='70')
        #plt.plot(cumcores9_sumfibrenumbers_60,cum_pc100umcores, '-o', color='magenta',label='60')
        ax.plot(cumcores9_sumfibrenumbers_50,cum_pc100umcores, '-o', color='orange',label='50')
        ax.plot(cumcores9_sumfibrenumbers_20,cum_pc100umcores, '-o', color='cyan',label='20')
        ax.set_xlim(0,5000)
        ax.set_ylim(0,100)
        ax.set_xlabel('Cumulative number of\n 105 micron fibres')
        ax.set_ylabel('Cumulative % galaxies\n imaged to 2Re')
        ax.text(3500, 60, "Num.\n bundles", color = 'black')
        #ax.text(7000, 50, "100", color = 'green')
        #ax.text(7000, 40, "90", color = 'cyan')
        #ax.text(7000, 30, "80", color = 'blue')
        #ax.text(7000, 20, "70", color = 'red')
        #ax.text(7000, 10, "60", color = 'magenta')
        ax.text(4000, 50, "50", color = 'orange')
        ax.text(4000, 40, "20", color = 'cyan')

        return fig, ax

    def make_Julias_plot(self, fig, axs):
        """ 
        Make an indetical plot to Julia's Hector sim code
        """
        if axs.shape != (3, 6):
            raise ValueError(f"Must pass axes of shape (3, 6)- these have shape {axs.shape}")

        self.plot_selected_sample_MZ_colour_Re(fig, axs[0, 0])
        self.plot_selected_sample_MZ(fig, axs[0, 1])
        self.plot_mass_hist(fig, axs[0, 2])
        self.plot_Re_hist(fig, axs[0, 3])
        self.plot_IFU_needed_for_2Re_hist(fig, axs[0, 4])
        self.plot_IFU_needed_for_2Re_kde(fig, axs[0, 5])

        self.plot_parent_sample(fig, axs[1, 0])
        self.plot_z_hist(fig, axs[1, 1])
        self.plot_ellipticity_hist(fig, axs[1, 2])
        self.plot_percentage_galaxy_sample_imaged_2Re(fig, axs[1, 3])
        self.cumulative_fraction_sample_imaged_to_2Re(fig, axs[1, 4])
        self.plot_SB_hist(fig, axs[1, 5])

        self.plot_SB_Re(fig, axs[2, 0])
        self.plot_gmi_hist(fig, axs[2, 1])
        self.plot_Mstar_z_gi_colour(fig, axs[2, 2])
        self.plot_z_Re(fig, axs[2, 3])
        self.plot_cumulative_fraction_imaged_vs_bundle_diameter(fig, axs[2, 4])
        self.cumulative_percentage_galaxies_imaged_2Re(fig, axs[2, 5])

        return fig, axs

if __name__ == '__main__':

    # get rid of the pandas false positive warning
    pd.options.mode.chained_assignment = None
    # Use the local style file I made
    plt.style.use('./HectorSim.mplstyle')

    import argparse
    parser = argparse.ArgumentParser(description="Simulate Hector Observations using a variety of input catalogues")
    parser.add_argument("BoundaryType", help="Must be Taipan or SAMI (not case sensitive)")
    parser.add_argument("zlimit", type=float, help="Maximum redshift")
    parser.add_argument("MstarMin", type=float, help="Minimum Mstar")
    parser.add_argument("MstarMax", type=float, help="Maximum Mstar")
    parser.add_argument("SparseFunction", type=int, help="Sparse selection function. Must be 1 or 2 (FIXME)")
    parser.add_argument("SourceCat", type=str, help="Short name for input catalogue")
    parser.add_argument("MSparseCut1", type=float, help="Select flat in M* above this mass")
    parser.add_argument("minRe", type=float, help="Minimum Re of galaxies to select (in kpc? arcsec?)")
    parser.add_argument("--quiet", "-q", action="store_true",
        help="Suppress printing to the terminal")
    parser.add_argument("--Dec_Min", type=float, default=np.nan, help="Minimum Dec of the observations. Useful for selecting a strip of targets")
    parser.add_argument("--Dec_Max", type=float, default=np.nan, help="Maximum Dec of the observations. Useful for selecting a strip of targets")



    args = parser.parse_args()


    config = _load_config(args.SourceCat)
    full_catalogue_table = _load_table(config, quiet=args.quiet)
    HectorSim_args = vars(args)


    full_catalogue_table['PRI_SAMI'] = 8
    full_catalogue_table['remaining_observations'] = 1
    low_mass_galaxies_for_repeats = (full_catalogue_table['Mstar'] < 9) & (full_catalogue_table['Mstar'] > 0)
    full_catalogue_table.loc[low_mass_galaxies_for_repeats, 'remaining_observations'] = 3


    HectorSim_args['entire_table'] = full_catalogue_table

    HS = HectorSim(**HectorSim_args)

    fig, axs = plt.subplots(nrows=3, ncols=6, figsize=(18.87, 5.94))
    HS.make_Julias_plot(fig, axs)
    fig.tight_layout()
    fig.subplots_adjust(hspace=0.4, wspace=0.28)

    fig_skyplot, ax_skyplot = plt.subplots(figsize=(18.87, 5.94))
    ax_skyplot.scatter(HS.entire_table.RA, HS.entire_table.DEC, c='k', s=1, alpha=0.5, label='Entire Input Catalogue')
    ax_skyplot.scatter(HS.possible_to_observe.RA, HS.possible_to_observe.DEC, c='red', s=1, alpha=0.5, label='Possible to observe (e.g. big/bright enough)')
    ax_skyplot.scatter(HS.selection_function.RA, HS.selection_function.DEC, c='blue', s=1, alpha=0.5, label='Above selection function')
    ax_skyplot.scatter(HS.selection_function_sparsely_sampled.RA, HS.selection_function_sparsely_sampled.DEC, c='magenta', s=2, label='Final catalogue (sparse sampling)')
    plt.show()
    #plt.savefig('53pc_below_0p6.png')
    #P.save_dataframe_as_FITS_table(HS.selection_function_sparsely_sampled, '/Users/samvaughan/Science/Hector/Tiling/TestingInputCatalogues/Tables/test_GAMA_Hector_Input_catalogue_53pc_below_0p6.fits')


    ## Work out the final source densities
    N_galaxy_targets = len(HS.selection_function_sparsely_sampled)
    N_hexabundles = 19
    final_source_density = N_galaxy_targets / HS.total_area # Galaxies per square degree
    source_density_per_Hector_field = final_source_density * np.pi * 1 * 1

    print(f"\n{len(HS.selection_function)} galaxies satisfy the selection function")
    print(f"Final Catalogue contains {N_galaxy_targets} galaxies")
    print(f"\nSource Density above selection function (blue points): {len(HS.selection_function)/HS.total_area * np.pi:.1f} galaxies per 2DF field")
    print(f"Source Density in final catalogue (magenta points): {source_density_per_Hector_field:.1f} galaxies per 2DF field")
    
    ## Print the fraction of galaxies below z=0.06
    fraction_below_0p6 = (HS.selection_function_sparsely_sampled.z < 0.06).sum()/len(HS.selection_function_sparsely_sampled)
    print(f"Fraction of galaixes below z=0.06: {fraction_below_0p6:.2f}")
    


    # fig, ax = plt.subplots()
    # HS.plot_parent_sample(fig, ax)






import numpy as np
import pandas as pd
import yaml
# For reading fits tables
from . import pandas_tools as P 
import os
import logging
from logging import handlers
from datetime import datetime
from pathlib import Path


def _load_config(filename):

    with open(filename, 'r') as ymlfile:
        config = yaml.safe_load(ymlfile)
    return config


def _read_table(fname):
    """
    Read in a table, either with fits or pandas
    """

    # Read in the whole table. May be an issue with massive tables?
    _, extension = os.path.splitext(os.path.expanduser(fname))

    fname = os.path.expanduser(fname)
    if extension == '.csv':
        table = pd.read_csv(fname)
    elif extension == '.tsv':
        table = pd.read_csv(fname, sep='\t')
    elif extension == '.fits':
        table = P.load_FITS_table_in_pandas(fname)
        #raise TypeError("Not yet written the code for fits... Exiting!")
    else:
        raise TypeError(f"Catalogue {fname} data type \
            not understood")
    return table


def create_output_directories(output_folder):

        # Make the folder if it doesn't exist
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Make the subdirectories if they don't exist
    subfolders_to_be_made = ['Logs', 'Configuration', 'Tiles', 'Plots', 'DistortionCorrected', "Allocation", "Allocation/tile_outputs", "Allocation/robot_outputs"]
    folders = dict(zip(subfolders_to_be_made, [Path(f'{output_folder}/{subfolder}') for subfolder in subfolders_to_be_made]))
    for name, p in folders.items():
        p.mkdir(parents=True, exist_ok=True)

    return folders

def set_up_loggers(config):
    """
    Set up a pair of loggers so we can save our progress to text files
    """

    output_folder = config['output_folder']
    output_filename_stem = config['output_filename_stem']

    now = datetime.now()
    current_time = now.strftime("%d_%m_%Y-%H_%M_%S")
    logger = logging.getLogger('tiling_progress')
    logger_R_code = logging.getLogger('R_code')
    logger.setLevel(logging.INFO)
    logger_R_code.setLevel(logging.INFO)


    formatter = logging.Formatter('%(asctime)s: - %(message)s')


    logger.handlers = []
    logger_R_code.handlers = []
    # This if statement is to avoid us having multiple loggers if we run the code twice in the same session. 
    if not logger.handlers:
        # Set up our console and file handlers
        console = logging.StreamHandler()
        console.setLevel(logging.INFO)

        file_handler = handlers.WatchedFileHandler(filename=f"{output_folder}/Logs/{output_filename_stem}_{current_time}.log")
        file_handler.setLevel(logging.INFO)

        console.setFormatter(formatter)
        file_handler.setFormatter(formatter)

        logger.addHandler(console)
        logger.addHandler(file_handler)

    # Now do the same for the R code log
    if not logger_R_code.handlers:
        file_handler_R_code = handlers.WatchedFileHandler(filename=f"{output_folder}/Logs/{output_filename_stem}_{current_time}_R_Output.log")

        file_handler_R_code.setLevel(logging.INFO)

        file_handler_R_code.setFormatter(formatter)
        logger_R_code.addHandler(file_handler_R_code)

    return logger, logger_R_code


import numpy as np
import scipy.spatial.distance as dist
import itertools
import matplotlib.pyplot as plt
import os
from operator import itemgetter
import warnings


def get_best_tile_centre_greedy(targets_df, outer_FOV_radius, inner_FoV_radius, n_xx_yy=100):
    """
    Given a set of x, y coordinates, find the position which would cover the most targets if we placed a field of view there.
    Inputs:
        targets_df (dataframe): A dataframe of target galaxies which have not yet been tiled. Must have columns 'RA' and 'DEC'
        n_xx_yy (int, optional): Number of grid points along each direction. The size of the grid is n_xx_yy**2, and needs to fit in memory in one go! So don't make this too large...
    """

    RA = targets_df.loc[:, 'RA']
    Dec = targets_df.loc[:, 'DEC']

    # Make the grid
    grid_coords = _get_grid(RA, Dec, n_xx_yy)

    n_targets_in_FOV = _calc_clashes(grid_coords, np.column_stack((RA, Dec)), proximity=outer_FOV_radius * 3600.0).sum(axis=1) - _calc_clashes(grid_coords, np.column_stack((RA, Dec)), proximity=inner_FoV_radius * 3600.0).sum(axis=1)

    # This will return the first position in the grid where n_targets_in_FOV is equal to its max value
    index = np.argmax(n_targets_in_FOV)

    return grid_coords[index, 0], grid_coords[index, 1]


def get_best_tile_centre_dengreedy(master_df, targets_df, outer_FOV_radius, inner_FoV_radius, n_xx_yy=100):
    """
    Given a set of x, y coordinates, find the position to place the tile according to Aaron's Dengreedy algorithm

    Inputs:
        master_df (dataframe): A dataframe of *all* target galaxies, irrespective of whether they've been tiled alread. Must have columns 'RA', 'DEC'
        targets_df (dataframe): A dataframe of target galaxies which have not yet been tiled. Must have columns 'RA', 'DEC'
        n_xx_yy (int, optional): Number of grid points along each direction. The size of the grid is n_xx_yy**2, and needs to fit in memory in one go! So don't make this too large...
    """

    RA_all = master_df.loc[:, 'RA']
    Dec_all = master_df.loc[:, 'DEC']

    RA_untiled = targets_df.loc[:, 'RA']
    Dec_untiled = targets_df.loc[:, 'DEC']

    grid_coords = _get_grid(RA_all, Dec_all, n_xx_yy)

    # See which tile centres have the largest number of clashes __outside the central 0.1*Radius of the field__. Julia says that the optics are poor here.

    n_targets_in_FOV_all = _calc_clashes(grid_coords, np.column_stack((RA_all, Dec_all)), proximity=outer_FOV_radius * 3600.0).sum(axis=1) - _calc_clashes(grid_coords, np.column_stack((RA_all, Dec_all)), proximity=inner_FoV_radius * 3600.0).sum(axis=1)
    n_targets_in_FOV_untiled = _calc_clashes(grid_coords, np.column_stack((RA_untiled, Dec_untiled)), proximity=outer_FOV_radius * 3600.0).sum(axis=1) - _calc_clashes(grid_coords, np.column_stack((RA_untiled, Dec_untiled)), proximity=inner_FoV_radius * 3600.0).sum(axis=1)

    ratio_of_targets = n_targets_in_FOV_untiled / n_targets_in_FOV_all
    ratio_of_targets[np.isnan(ratio_of_targets)] = 0.0

    # If there are many places we can put a tile, randomise the location
    if np.sum(ratio_of_targets == np.max(ratio_of_targets)):
        index = np.random.choice(np.where(ratio_of_targets == np.max(ratio_of_targets))[0])
    else:
        index = np.argmax(ratio_of_targets)

    return grid_coords[index, 0], grid_coords[index, 1]


def _get_grid(RA, Dec, n_xx_yy):
    """
    Make a grid of points which encloses a set of given RAs and Decs
    Inputs:
        RA (array): A list of RA values of targets
        Dec (array): A list of Dec values of targets
        n_xx_yy (int): Number of grid points along **one** axis. Final grid has n_xx_yy^2 points!
    """
    xx = np.linspace(RA.min() - 0.5, RA.max() + 0.5, n_xx_yy)
    yy = np.linspace(Dec.min() - 0.5, Dec.max() + 0.5, n_xx_yy)

    # This is a bit sad. array(list(generator)).
    grid_coords = np.array(list(itertools.product(xx, yy)))

    return grid_coords


def find_nearest(x, y, grid):
    """
    Find the nearest grid point to each target (at coordinates (x, y))
    Inputs:
        x (array_like): A list of target x coordinates
        y (array_like): A list of target y coordinates
        grid (2D array): A two dimensional array of grid coordinates to test against. Columns are x (grid[:, 0]) and y (grid[:, 1])
    """

    if len(x) != len(y):
        raise ValueError(f"x and y must be the same length: Currently {len(x)} and {len(y)}")
    if grid.ndim != 2:
        raise ValueError("grid must have two dimensions")

    nearest_points = np.zeros(len(x))
    for i, (target_x, target_y) in enumerate(zip(x, y)):
        nearest_points[i] = np.argmin(np.sqrt((grid[:, 0] - target_x)**2 + (grid[:, 1] - target_y)**2))
    return nearest_points


def check_if_in_fov(df, xcen, ycen, inner_radius, outer_radius):
    """
    Return a binary mask if points are within a circular field of view of radius R
    Inputs:
        df (dataframe): A dataframe of targets. Must have columns 'RA' and 'DEC'
        xcen (float): x (or RA) coordinate of centre
        ycen (float): y (or DEC) coordinate of centre
        radius (float): radius of circle in degrees
    """

    x = df.RA
    y = df.DEC

    return (np.sqrt((x - xcen)**2 + (y - ycen)**2) < outer_radius) & (np.sqrt((x - xcen)**2 + (y - ycen)**2) > inner_radius)


def find_clashes(df1, df2, proximity):
    """
    Given some proximity, find objects in two dataframes which will clash
    Inputs:
        df1 (dataframe): Dataframe of targets. Must have columns 'RA' and 'DEC'
        df2 (dataframe): Dataframe of targets. Must have columns 'RA' and 'DEC'
        proximity (float): Smallest distance which two objects can come before clashing. Measured in arcseconds

    """

    if proximity < 0:
        raise ValueError("Proximity must be positive")

    # Make sure that both the input coordinate lists are 2D
    XA = np.atleast_2d(df1.loc[:, ['RA', 'DEC']].values)
    XB = np.atleast_2d(df2.loc[:, ['RA', 'DEC']].values)

    clashes = _calc_clashes(XA, XB, proximity)

    return clashes


def _calc_clashes(XA, XB, proximity):
    """
    Given two lists of coordinates, return a boolean mask highlighting only those which clash within some minimum proximity. Note that we ignore self clashes by ignoring clashes with a distance = 0.0
    """
    distance_matrix = dist.cdist(XA, XB)
    clashes = (distance_matrix > 0.0) & (distance_matrix < proximity / 3600.0)

    return clashes


def noclash(df1, df2, proximity):
    """
    Take two dataframes, find those which are within some proximity of each other and then return the first dataframe as is, the SECOND dataframe without the clashing elements, and the number of elements which clash. Useful for selecting guide stars which don't clash with targets in a tile.
    Inputs:
        df1 (dataframe): Dataframe of targets. Must have columns 'RA' and 'DEC'
        df2 (dataframe): Dataframe of targets. Must have columns 'RA' and 'DEC'
        proximity (float): Smallest distance which two objects can come before clashing. Measured in arcseconds
    """

    clashes = find_clashes(df1, df2, proximity)
    ncollide = np.sum(clashes)

    # If a target has a clash with another, there'll be a true in its clashes row. Check for this and only pick things with no clash
    good_indices_1 = clashes.sum(axis=0) == 0

    if ncollide == 0:
        return df1, df2, ncollide
    else:
        return df1, df2.iloc[good_indices_1], ncollide


def select_stars_for_tile(star_df, tile_df, proximity, Nsel, star_type):
    """
    Given a tile of targets, select Nsel worth of stars by selecting stars which don't clash with any of our targets. The returned stars are sorted by either their priority (for standard stars) or their R-band magnitude (for guide stars)

    Inputs:
        star_df (dataframe): a dataframe of stars (guide or standard). Must have columns "RA", "DEC" and either "priority" or "R_MAG_AUTO" (see below)
        tile_df (dataframe): a dataframe containing Nsel targets, created from using 'unpick'. Must have columns "RA" amd "DEC"
        proximity (float): distance between two adjacent bundles, in arcseconds.
        Nsel (int): Return this many stars for each tile.
        star_type (string): One of either "standards" or "guides". Selects whether we sort the star dataframe by "priority" (for standards) or "R_MAG_AUTO" (for guides)
    """
    if star_type == 'standards':
        star_df = star_df.sort_values(by='priority', ascending=False)
    elif star_type == 'guides':
        star_df = star_df.sort_values(by='R_MAG_AUTO', ascending=True)
    else:
        raise NameError(f"star_type must be either 'standards' or 'guides'; currently {star_type}")

    tile_df, non_clashing_stars, _ = noclash(tile_df, star_df, proximity)

    # Make sure the non clasing stars also don't clash with themselves 
    _, non_clashing_stars, _ = noclash(non_clashing_stars, non_clashing_stars, proximity)

    if Nsel > len(non_clashing_stars):
        Nsel = len(non_clashing_stars)

    return non_clashing_stars.iloc[:Nsel]


def select_targets(all_targets_df, proximity, Nsel, selection_type='most_clashing', fill_spares_with_repeats=False):
    """
    Given a dataframe of targets, select Nsel galaxies to observe. These can't be nearer to each other than 'proximity'. Return a dataframe of just the new tile we've made.
    Inputs:
        df (dataframe): A dataframe of targets. Must have columns "RA", "DEC", "PRI_SAMI" (i.e. priority), "ALREADY_TILED" (i.e. tiled before this current iteration).
        proximity (float): Smallest distance between two targets (in arcseconds)
        Nsel (int): Number of targets to select in each tile
    """

    # Select things which haven't been already tiled or selected as a target in this iteration
    target_mask = (all_targets_df['ALREADY_TILED'] == False)
    df = all_targets_df.loc[target_mask]
    already_tiled_in_FOV = all_targets_df.loc[~target_mask]

    # Get the priority of each target
    priorities = df['PRI_SAMI'].values

    # Find which of these targets clash with each other
    clashes = find_clashes(df, df, proximity)
    # n_clashes = clashes.sum(axis=0)

    targets = []
    isel_values = []
    untiled_galaxies = np.arange(len(df))  # [sorted_by_pri_then_nclashes]
    # priorities = priorities[sorted_by_pri_then_nclashes]

    # Now go through and add the target which clashes with the most things to the tile.
    # Keep doing this until we have selected 'Nsel' things or we run out of targets
    while (len(targets) < Nsel) and (len(untiled_galaxies) > 0):

        # Find any clashes in the targets
        n_clashes = np.sum(clashes, axis=0)[untiled_galaxies]
        # This is a dictionary of the priority of every target and then the number of things it clashes with
        clash_and_priority_dict = dict(zip(untiled_galaxies, zip(priorities[untiled_galaxies], n_clashes)))

        # Sort by priority then by number of clashes. Important that the tuples in clash_and_priority_dict are in the order (priority, n_clashes)
        galaxies_in_order = sorted(clash_and_priority_dict.items(), key=itemgetter(1), reverse=True)

        # If we have a clash...
        if np.sum(n_clashes) > 0:

            # Now find the target to select
            if selection_type == 'most_clashing':
                selected_target = galaxies_in_order[0][0]
            elif selection_type == 'random':
                random_choice = np.random.randint(low=0, high=len(galaxies_in_order))
                selected_target = galaxies_in_order[random_choice][0]
            else:
                raise NameError('Unknown selection type')
            # And find the targets it clashes with
            things_it_clashes_with = np.where(clashes[selected_target])[0]
            # Tile the most clashing target
            targets.append(selected_target)

            # Add its 'isel' value as its priority plus one, since it's a clashing target
            isel_values.append(priorities[selected_target] + 1)

            # ... and remove the things this target clashes with from the list of targets in this tile
            untiled_galaxies = [u for u in untiled_galaxies if u not in things_it_clashes_with]
            untiled_galaxies.remove(selected_target)

        # Else if there aren't any clashes remaining...
        else:
            # Just add everything left
            N_to_append = min(Nsel - len(targets), len(untiled_galaxies))
            # random_choice = np.random.choice(np.arange(len(untiled_galaxies)), replace=False, size=N_to_append)

            unclashing_targets = [galaxies_in_order[r][0] for r in range(N_to_append)]

            targets.extend(unclashing_targets)
            isel_values.extend(priorities[unclashing_targets])

            # Remove things we've tiled
            untiled_galaxies = [u for u in untiled_galaxies if u not in targets]

    tile_df = df.iloc[targets]

    if fill_spares_with_repeats:
        if len(tile_df) < Nsel:
            # print("!!")
            # We now see if any things which have already been tiled can be repeated
            # We check to see whether everything in the Field of View clashes with anything which has already been tiled.
            # NOTE that this a quick-fix way of doing things- for example, if two things are in a close pair, using this method they will never be added to the tile as a repeat, since they will always clash with each other. The correct thing to do would be to add in things to repeat one by one, as in the method above
            # Also need to fix things to pick 'high priority' targets to repeat preferentially
            clashes = find_clashes(tile_df.append(already_tiled_in_FOV), already_tiled_in_FOV, proximity)
            n_clashes = clashes.sum(axis=0)
            if len(already_tiled_in_FOV) > 0 & (np.any(n_clashes == 0)):
                unclashing_repeats = np.where(n_clashes == 0)[0]
                N_to_append = min(Nsel - len(tile_df), len(unclashing_repeats))
                repeats_to_fill_hexabundles = np.random.choice(unclashing_repeats, N_to_append, replace=False)
                tile_df = tile_df.append(already_tiled_in_FOV.iloc[repeats_to_fill_hexabundles])
                tile_df.loc[tile_df.ALREADY_TILED == True, 'isel'] = 1.0
                isel_values.extend([1.0] * len(repeats_to_fill_hexabundles))

                if len(tile_df) < Nsel:
                    print(f"Can only select {len(targets)} new targets for this field. Can only select {N_to_append} repeats. Tile length is only {len(tile_df)}!!!")
                else:
                    print(f"Can only select {len(targets)} new targets for this field. Select {N_to_append} repeats. Tile length is {len(tile_df)}")
            else:
                print(f"Can only select {len(targets)} new targets for this field. Can't select any repeats. Tile length is only {len(tile_df)}!!!")

    # # # If we don't fill a tile, append already observed galaxies till we get to the right number
    # # # These are set an isel value of 1.
    # # However, check there are other things left to tile! Otherwise we add repeats of the same targets which breaks things
    # if (len(tile_df) < Nsel) & (len(all_targets_df) > len(tile_df)):
        
    #     N_missing = Nsel - len(tile_df)
        
    #     gals_added = 0
    #     # Loop through and add targets, checking each time if they clash with the tile
    #     # This is a bit slower than the proper way of doing things above (since we're calling find_clashes many times) but the tile should be small enough that it's not a big deal. 
    #     for i in range(N_missing):

    #         # We're now finding clashes between the tile and *everything* in the field, not just things we haven't observed
    #         targets_not_already_in_tile = all_targets_df[~all_targets_df.index.isin(tile_df.index)]
    #         clashes = find_clashes(targets_not_already_in_tile, tile_df, proximity=proximity)
    #         n_clashes = clashes.sum(axis=1)
    #         if np.any(np.atleast_1d(n_clashes) == 0):
    #             unclashing_index = np.random.choice(np.where(n_clashes == 0)[0])
    #             gals_added += 1
    #             tile_df = tile_df.append(all_targets_df.iloc[unclashing_index])
    #         else:
    #             break

    #     isel_values.extend([1] * gals_added)
    

    if len(tile_df) < Nsel:
        if (len(all_targets_df) > len(tile_df)):
            print(f"Can't select {Nsel} targets for this field. Info: {len(all_targets_df)} in FOV and {len(tile_df)} in tile!")
        else:
            targets_not_already_in_tile = all_targets_df[~all_targets_df.index.isin(tile_df.index)]
            clashes = find_clashes(targets_not_already_in_tile, tile_df, proximity=proximity)
            n_clashes = clashes.sum(axis=1)
            print(f"Can't select {Nsel} targets for this field. Info: {len(all_targets_df)} in FOV; {len(tile_df)} in tile; {len(targets_not_already_in_tile)} are in FOV but not tiled, of which each one clashes {n_clashes} times")

    # Check if we clash with ourself- this should never happen
    clashes = find_clashes(tile_df, tile_df, proximity)
    if np.sum(clashes) != 0:
        raise ValueError("Our tile seems to clash with itself... This should never happen!")

    return tile_df, isel_values


def make_best_tile(df_targets, df_guide_stars, df_standard_stars, tiling_parameters, tiling_type, selection_type='most_clashing', fill_spares_with_repeats=False):
    """
    Put all the above functions togther and make a tile. Note that this function __doesn't__ update any tiling flags in the overall database. This should be done afterwards, so that we can integrate things with the Hector configuration code- the 19 best targets we pick might not actually be tile-able, so we don't want to mark things as tiled if the config code needs to select backups.
    Inputs:
        df_targets (dataframe): A dataframe of galaxy targets. Must include columns 'RA', 'DEC', "PRI_SAMI" and "TILED"
        df_guide_stars (dataframe): A dataframe of galaxy targets. Must include columns 'RA', 'DEC' and "R_MAG_AUTO"
        df_standard_stars (dataframe): A dataframe of standard stars. Must include columns 'RA', 'DEC' and 'priority'
        tiling_parameters (dict): A dictionary containing the various parameters to do with the tiling. So far, necessary keys are 'Hector_FOV_radius', 'proximity', 'Nsel', 'Nsel_guides' and 'Nsel_standards'
    Returns:
        A tuple of:
            * The original df_targets dataframe
            * A dataframe containing the targets for this tile
            * A dataframe containing the guides for this tile
            * A dataframe containing the standard stars for this tile
            * The tile RA
            * The tile Dec
    """

    Hector_FOV_outer_radius = tiling_parameters['Hector_FOV_outer_radius']
    Hector_FOV_inner_radius = tiling_parameters['Hector_FOV_inner_radius']
    proximity = tiling_parameters['proximity']
    Nsel = tiling_parameters['Nsel']
    Nsel_guides = tiling_parameters['Nsel_guides']
    Nsel_standards = tiling_parameters['Nsel_standards']

    # Do some basic error checking
    if Hector_FOV_inner_radius < 0:
        raise ValueError(f"FoV Inner radius must be positive! Currently it's {Hector_FOV_inner_radius}")
    if Hector_FOV_outer_radius < 0:
        raise ValueError(f"FoV Outer radius must be positive! Currently it's {Hector_FOV_outer_radius}")
    if Hector_FOV_outer_radius < Hector_FOV_inner_radius:
        raise ValueError(f"Outer radius must be greater than inner radius! Currently O.R. is {Hector_FOV_outer_radius} and I.R. is {Hector_FOV_inner_radius}")

    untiled = df_targets['ALREADY_TILED'] == False # & (df_targets['PRI_SAMI'] == 8)

    # FIXME
    # This will cause an issue if the separation of two grid points is larger than the FOV size
    n_xx_yy = max(min(500, int(untiled.sum() / 10)), 50)

    if tiling_type == 'greedy':
        tile_RA, tile_Dec = get_best_tile_centre_greedy(df_targets.loc[untiled], outer_FOV_radius=Hector_FOV_outer_radius, inner_FoV_radius=Hector_FOV_inner_radius, n_xx_yy=n_xx_yy)
    elif tiling_type == 'dengreedy':
        tile_RA, tile_Dec = get_best_tile_centre_dengreedy(df_targets, df_targets.loc[untiled], outer_FOV_radius=Hector_FOV_outer_radius, inner_FoV_radius=Hector_FOV_inner_radius, n_xx_yy=n_xx_yy)
    else:
        raise NameError("tiling_type must be one of 'greedy' or 'dengreedy'")

    # Only select targets, guides and standards within the FoV
    # check_if_in_FOV returns a boolean mask
    inner_targets = df_targets.loc[check_if_in_fov(df_targets, tile_RA, tile_Dec, outer_radius=Hector_FOV_outer_radius, inner_radius=Hector_FOV_inner_radius), :]
    inner_guides = df_guide_stars.loc[check_if_in_fov(df_guide_stars, tile_RA, tile_Dec, outer_radius=Hector_FOV_outer_radius, inner_radius=Hector_FOV_inner_radius), :]
    inner_standards = df_standard_stars.loc[check_if_in_fov(df_standard_stars, tile_RA, tile_Dec, outer_radius=Hector_FOV_outer_radius, inner_radius=Hector_FOV_inner_radius), :]

    # Select which targets to keep
    tile_members, isel_values = select_targets(inner_targets, proximity, Nsel, selection_type=selection_type, fill_spares_with_repeats=fill_spares_with_repeats)

    # Add the 'isel' values to the main dataframe
    df_targets.loc[tile_members.index, 'isel'] = isel_values
    tile_members.loc[:, 'isel'] = isel_values

    # Now add in the guide stars and standard stars
    # Make sure none clash with any of the targets in that tile
    guide_stars_for_tile = select_stars_for_tile(inner_guides, tile_members, proximity=proximity, Nsel=Nsel_guides, star_type='guides')
    standard_stars_for_tile = select_stars_for_tile(inner_standards, tile_members, proximity=proximity, Nsel=Nsel_standards, star_type='standards')

    standard_stars_for_tile['isel'] = standard_stars_for_tile['priority'] + 10

    return df_targets, tile_members, guide_stars_for_tile, standard_stars_for_tile, tile_RA, tile_Dec


def save_tile_outputs(outfolder, df_targets, tile_df, guide_stars_for_tile, standard_stars_for_tile, tile_RA, tile_Dec, tiling_parameters, tile_number, plot=True):
    """
    Save the outputs from a single tile. These are:
        * A text file called tile_{i}.fld, which contains things to be observed the Hector science bundles. This has columns CATID, RA, DEC, mag, type (where type is 1 for a galaxy target and 0 for a standard star) and isel (which is a bit like the priority they should be targeted in). The targets are sorted by priority!
        * A text file called guide_tile_{i}.fld, which contains guide stars (observed with the Hector guide bundles). This has columns RA, DEC, mag
        * A plot of the field with the Nsel best targets selected by this code. Note that this may not resemble the final tile selected by the configuration code!
    These outputs will be bundled together in a single outfolder" folder, which contains subfolders 'Tiles' and 'Plots'
    Inputs:
        outfolder (str): Location for the output files
        df_targets (dataframe): Dataframe of all targets. This is used for plotting
        tile_df (dataframe): Dataframe of targets for a single tile
        guide_stars_for_tile (dataframe): Dataframe of guide stars for a single tile
        standard_stars_for_tile (dataframe): Dataframe of standard stars for a single tile
        tile_RA (float): RA of tile centre
        tile_DEC (float): DEC of tile centre
        tiling parameters (dict): Dictionary of tiling parameters
        tile_number (int): Number of the tile
        plot (bool, optional): Whether to save a plot of the tile. Default is True
    """

    tile_out_name = f"tile_{tile_number:03}.fld"
    guide_out_name = f"guide_tile_{tile_number:03}.fld"

    save_tile_text_file(outfolder, tile_out_name, tile_df, standard_stars_for_tile, tile_RA, tile_Dec, tiling_parameters)
    save_guide_text_file(outfolder, guide_out_name, guide_stars_for_tile, tile_RA, tile_Dec, tiling_parameters)

    if plot:
        if not os.path.exists(f'{outfolder}/Plots'):
            os.makedirs(f'{outfolder}/Plots')

        fig, ax = plot_tile(tile_df, guide_stars_for_tile, standard_stars_for_tile, df_targets, tile_RA, tile_Dec, tiling_parameters['Hector_FOV_outer_radius'], tiling_parameters['Hector_FOV_inner_radius'], tile_number, tiling_parameters['proximity'])
        fig.savefig(f'{outfolder}/Plots/tile_{tile_number:03}.png')
        plt.close('all')

    return 0


def plot_survey_completeness_and_tile_positions(tile_positions, df_targets, tiling_parameters, fig=None, ax=None, completion_fraction_to_calculate=0.95, verbose=True):
    """
    Make an overall plot of the survey completeness and the positions of each tile.
    Inputs:
        tile_positions (list): A two component list. First element is a list of RA values of the tile centres, second component is a list of Dec values of the tile centres.
        tiling_parameters (dict): A dictionary of tiling parameters. Must have  a 'Hector_FOV_radius' key
    """

    tile_RAs, tile_DECs = tile_positions

    fig, axs = plt.subplots(nrows=1, ncols=2, figsize=(17, 5))
    axs[0].axis('equal')
    
    tiles = df_targets.groupby('Tile_number').groups

    for tile_number in tiles:
        members = df_targets.loc[tiles[tile_number]] # This selects by index, NOT integer position.
        axs[0].scatter(members.RA, members.DEC) 
        tile_footprint = plt.Circle(xy=(tile_RAs[tile_number], tile_DECs[tile_number]), radius=tiling_parameters['Hector_FOV_outer_radius'], facecolor='None', edgecolor='k')
        axs[0].add_artist(tile_footprint)

    
    completeness, completion_fraction_to_calculate, used_tiles_to_get_to_x, minimum_number_of_tiles_for_x, efficiency_xpc, efficiency = calculate_completeness_stats(df_targets, tiling_parameters['N_targets_per_Hector_field'], completion_fraction_to_calculate=completion_fraction_to_calculate, verbose=verbose)

    # Plot the completeness as a function of tile number
    xx = np.arange(1, len(completeness) + 1)
    axs[1].plot(xx, tiling_parameters['N_targets_per_Hector_field'] / len(df_targets) * xx, c='k', linestyle='dashed')
    axs[1].plot(xx, completeness, c='r', zorder=10)
    axs[1].axhline(completion_fraction_to_calculate, c='0.8', linestyle='dashed')
    axs[1].axhline(1, c='0.5')
    axs[1].set_ylim(0, 1.2)
    axs[1].axvline(used_tiles_to_get_to_x, c='0.8', linestyle='dashed')
    axs[1].axvline(minimum_number_of_tiles_for_x, c='0.8', linestyle='dashed')
    axs[1].set_xlabel(r'$N_{\rm{tiles}}$')
    axs[1].set_ylabel('Completeness')

    return fig, axs


def _calc_completeness(df_targets):
    """
    Given a tiling dataframe, work out the fractional completeness of the tiling after each tile.
    Inputs:
        df_targets (dataframe): a dataframe with a row for each target. Must have a column 'Tile_number'
    """

    completeness = np.cumsum([np.sum(df_targets['Tile_number'] == i) / len(df_targets) for i in np.unique(df_targets['Tile_number'])])

    return completeness


def calculate_completeness_stats(df_targets, N_targets_per_Hector_field, completion_fraction_to_calculate=0.95, verbose=True):
    """
    Given a set of tiles, calculate some stats about the efficiency to get to a given completeness fraction
    Inputs:
        df_targets (dataframe): a dataframe with a row for each target. Must have a column 'Tile_number'
        N_targets_per_Hector_field (dict): The numnber of hexabundles we can place on galaxy targets
        completeness_fraction_to_calculate (float, default=0.95): calculate the efficiency to reach this completeness fraction. This is defined as actual number of tiles used / minimum number of tiles possible) 
        verbose (bool, default=True): print efficiency stats or not.  
    """

    if (completion_fraction_to_calculate > 1) or (completion_fraction_to_calculate < 0):
        raise ValueError(f"completion_fraction_to_calculate must be between 0 and 1: currently {completion_fraction_to_calculate}")

    # Work out the completeness after each tile
    completeness = _calc_completeness(df_targets)

    used_tiles = df_targets['Tile_number'].max() + 1  # Since we start indexing from 0
    minimum_number_of_tiles = np.ceil(len(df_targets) / N_targets_per_Hector_field)
    efficiency = minimum_number_of_tiles / used_tiles

    used_tiles_to_get_to_x = np.where(completeness > completion_fraction_to_calculate)[0][0] + 1
    minimum_number_of_tiles_for_x = np.ceil(completion_fraction_to_calculate * len(df_targets) / N_targets_per_Hector_field)
    efficiency_xpc = minimum_number_of_tiles_for_x / used_tiles_to_get_to_x

    if verbose:
        print(f"Efficiency for completion={completion_fraction_to_calculate}: {100*efficiency_xpc:.3f}%")
        print(f"Efficiency for completion=1: {100*efficiency:.3f}%")

    return completeness, completion_fraction_to_calculate, used_tiles_to_get_to_x, minimum_number_of_tiles_for_x, efficiency_xpc, efficiency


def plot_tile(tile_df, guide_df, standards_df, catalogue_df, tile_RA, tile_Dec, tile_outer_radius, tile_inner_radius, tile_number, proximity, fig=None, ax=None):
    """
    Make a plot of an individual tile.
    Inputs:
        tile_df (dataframe): Dataframe of a tile to plot
        guide_df (dataframe): Dataframe of guide stars from this tile
        standards_df (dataframe): Dataframe of standard stars from this tile
        catalogue_df (dataframe): Dataframe of the entire field, to plot as background
        tile_RA (float): RA of the centre of this tile
        tile_Dec (float): Dec of the centre of this tile
        tile_outer_radius (float): Outer Radius of the FoV (in degrees)
        tile_inner_radius (float): Inner Radius of the FoV (in degrees)
        tile_number (int): Number of the tile
        proximity (float): Radius of the tile
        fig, ax (optional): An exisiting Figure/Axis object to plot on
    """
    top_target_mask = catalogue_df['CATAID'].isin(tile_df['CATAID']) 

    if fig is None or ax is None:
        fig, ax = plt.subplots(figsize=(14, 10))

    ax.scatter(catalogue_df.RA, catalogue_df.DEC, marker=',', s=1, c='0.5')
    ax.scatter(tile_df.RA, tile_df.DEC, marker='x', c='k', s=30, label='All Targets', zorder=10)
    ax.scatter(catalogue_df.loc[top_target_mask, 'RA'], catalogue_df.loc[top_target_mask, 'DEC'], marker='x', c='r', s=50, label='Top Targets', zorder=10)
    ax.scatter(standards_df.RA, standards_df.DEC, marker='o', c='g', s=20, label='Standards')
    ax.scatter(guide_df.RA, guide_df.DEC, marker='s', c='b', s=20, label='Guides')
    ax.scatter(tile_RA, tile_Dec, marker='+', c='k', s=200, label='Tile Centre')


    for index, row in tile_df.iterrows():

        circle = plt.Circle(xy=(row.RA, row.DEC), radius=proximity / 3600.0, facecolor='None', edgecolor='k', linestyle='dashed')
        ax.add_artist(circle)


    ax.set_aspect('equal')
    ax.set_xlim(tile_RA - 1.1 * tile_outer_radius, tile_RA + 1.1 * tile_outer_radius)
    ax.set_ylim(tile_Dec - 1.1 * tile_outer_radius, tile_Dec + 1.1 * tile_outer_radius)

    circle = plt.Circle((tile_RA, tile_Dec), tile_outer_radius, facecolor='None', edgecolor='r')
    ax.add_artist(circle)
    # Add the total 2DF field of view in black
    circle = plt.Circle((tile_RA, tile_Dec), 1.0, facecolor='None', edgecolor='k')
    ax.add_artist(circle)

    # Add the central region which we're ignoring
    circle = plt.Circle((tile_RA, tile_Dec), tile_inner_radius, facecolor='0.1', edgecolor='0.1', alpha=0.3)
    ax.add_artist(circle)


    ax.legend(bbox_to_anchor=(1, 0.5))

    ax.set_xlabel('Right Ascension')
    ax.set_ylabel('Declination')

    ax.set_title(f'Tile {tile_number}')

    return fig, ax


def save_tile_text_file(outfolder, out_name, tile_df, standard_stars_for_tile, tile_RA, tile_Dec, tiling_parameters):
    """
    Save a text file of things to be observed with Hector science bundles (i.e. targets and standards). The second line of this text file __must__ be the tile RA and DEC seperated by a space.
    """

    if not os.path.exists(f"{outfolder}/Tiles"):
        os.makedirs(f"{outfolder}/Tiles")

    # Make sure the tile_df is sorted by priority
    tile_df = tile_df.sort_values(by='isel', ascending=False)

    # Rename the column headings so things match
    targets_renamer = dict(CATAID="ID", r_mag='mag')
    tile_df = tile_df.rename(columns=targets_renamer)
    tile_df['type'] = 1

    standards_renamer = dict(CoADD_ID='ID', R_MAG_AUTO='mag')
    standard_stars_for_tile = standard_stars_for_tile.rename(columns=standards_renamer)
    standard_stars_for_tile['type'] = 0

    # combined_stars_targets_df = tile_df[['ID', 'RA', 'DEC', 'mag', 'type', 'isel']].append(standard_stars_for_tile[['ID', 'RA', 'DEC', 'mag', 'type', 'isel']])
    combined_stars_targets_df = tile_df.append(standard_stars_for_tile, sort=True)[['ID', 'RA', 'DEC', 'Re', 'Mstar', 'z', 'GAL_MAG_G', 'GAL_MAG_I', 'GAL_MU_0_G', 'GAL_MU_0_I', 'GAL_MU_0_R', 'GAL_MU_0_U',
       'GAL_MU_0_Z', 'GAL_MU_E_G', 'GAL_MU_E_I', 'GAL_MU_E_R', 'GAL_MU_E_U',
       'GAL_MU_E_Z', 'GAL_MU_R_at_2Re', 'GAL_MU_R_at_3Re', 'Dingoflag', 'Ellipticity_r',
       'IFU_diam_2Re', 'MassHIpred',  'PRI_SAMI',
       'SersicIndex_r', 'WALLABYflag', 'g_m_i', 'isel', 'mag',
       'priority', 'remaining_observations', 'Tile_number', 'ALREADY_TILED', 'type']]

    # Write a CSV file with the header we want
    with open(f"{outfolder}/Tiles/{out_name}", 'w') as f:
        f.write("# Target and Standard Star file from Sam's tiling code\n")
        f.write(f"# {tile_RA} {tile_Dec}\n")
        f.write(f"# Proximity Value: {tiling_parameters['proximity']}\n")
    combined_stars_targets_df.to_csv(f"{outfolder}/Tiles/{out_name}", sep=',', mode='a', index=False)

    return 0


def save_guide_text_file(outfolder, out_name, guide_stars_for_tile, tile_RA, tile_Dec, tiling_parameters):
    """
    Save a text file of things to be observed with Hector guide bundles (i.e. guides). The second line of this text file __must__ be the tile RA and DEC seperated by a space.
    """

    if not os.path.exists(f"{outfolder}/Tiles"):
        os.makedirs(f"{outfolder}/Tiles")

    # Make sure the tile_df is sorted by priority
    guide_stars_for_tile = guide_stars_for_tile.sort_values(by='R_MAG_AUTO')

    guides_renamer = dict(CoADD_ID='ID', R_MAG_AUTO='mag')
    guide_stars_for_tile = guide_stars_for_tile.rename(columns=guides_renamer)
    guide_stars_for_tile['type'] = 2

    # Write a CSV file with the header we want
    with open(f"{outfolder}/Tiles/{out_name}", 'w') as f:
        f.write("# Target and Standard Star file from Sam's tiling code\n")
        f.write(f"# {tile_RA} {tile_Dec}\n")
        f.write(f"# Proximity Value: {tiling_parameters['proximity']}\n")

    guide_stars_for_tile[['ID', 'RA', 'DEC', 'mag', 'type']].to_csv(f"{outfolder}/Tiles/{out_name}", sep=' ', mode='a', index=False)

    return 0

############################################################################################################



import numpy as np
import string
import sys
import re
from ..general_operations.trigonometry import convert_radians_to_degrees


# assigning magnet labels based on type
def assign_magnet_labels(magnet, index1):
    # adding labels for rectangular magnet
    if magnet.__class__.__name__ == 'rectangular_magnet':
        magnet.magnet_label = 'R%02d' % (index1)
        index1 += 1

    # adding labels for circular magnet
    elif magnet.__class__.__name__ == 'circular_magnet':
        if magnet.rads < 0.396:
            magnet.magnet_label = 'Blu'
        elif 0.396 <= magnet.rads < 0.627:
            magnet.magnet_label = 'Gre'
        elif 0.627 <= magnet.rads < 0.823:
            magnet.magnet_label = 'Yel'
        elif magnet.rads >= 0.823:
            magnet.magnet_label = 'Mag'

    return magnet, index1


# flagging unresolvable conflicts by recording them in file and assigning '0' placement order to them
def check_for_unresolvable_conflict(magnet, conflictFile, max_order, clusterNum, tileNum):
    # placement index checking and assigning placement order accordingly
    if magnet.placement_index == None:
        order = 0
        with open(conflictFile, 'a') as fp:
            conflict = magnet.__class__.__name__ + ' ' + str(int(magnet.index)) + ' is an unresolvable conflict,' + \
                       ' detected in Cluster ' + str(clusterNum) + ', tile ' + str(tileNum) + '\n'
            fp.write(conflict)

    else:
        order = 1
        for increment in range(max_order + 1):
            if magnet.placement_index == max_order - increment:
                order += increment

    return order


# when having more than 1 placement orders, the magnets with 1st placement order are to be assigned any pickup direction
def add_pickupDirection(magnet, available_pickup):
    # magnets with placement order of 1 are placed at first so can be placed from any pickup direction
    if available_pickup == []:
        if magnet.__class__.__name__ == 'circular_magnet':
            available_pickup = ['TR', 'TL', 'RI', 'RO']
        elif magnet.__class__.__name__ == 'rectangular_magnet':
            available_pickup = ['I', 'O']

    return magnet, available_pickup


# prioritization of pickup directions when more than one option is available for a magnet
def assign_preferable_pickupDirection(available_pickup):

    if 'TR' in available_pickup:
        available_pickup = []
        available_pickup = 'TR'
    elif 'TL' in available_pickup:
        available_pickup = []
        available_pickup = 'TL'
    elif 'RO' in available_pickup:
        available_pickup = []
        available_pickup = 'RO'
    elif 'RI' in available_pickup:
        available_pickup = []
        available_pickup = 'RI'
    elif 'I' in available_pickup:
        available_pickup = []
        available_pickup = 'I'
    elif 'O' in available_pickup:  ### REMEMBER TO CHANGE Inwards to Toward and Outwards to Away ###
        available_pickup = []
        available_pickup = 'O'

    return available_pickup


# angles required by the robotic arm to pick up and then put down the magnet in correct position
def calculate_pickup_putdown_angles(magnet, available_pickup):
    azAngs = convert_radians_to_degrees(magnet.azAngs)

    # rotation for pickup and putdown of circular magnets
    if magnet.__class__.__name__ == 'circular_magnet':
        if 'TR' in available_pickup:
            magnet.rotation_pickup = 90
            if azAngs < 90:
                magnet.rotation_putdown = 90 - azAngs
            elif azAngs >= 90:
                magnet.rotation_putdown = 360 + (90 - azAngs)
        elif 'TL' in available_pickup:
            magnet.rotation_pickup = 270
            if azAngs < 90:
                magnet.rotation_putdown = 270 - azAngs
            elif azAngs >= 90:
                magnet.rotation_putdown = 270 - azAngs
        elif 'RO' in available_pickup:
            magnet.rotation_pickup = 180
            if azAngs < 90:
                magnet.rotation_putdown = 180 - azAngs
            elif azAngs >= 90:
                magnet.rotation_putdown = 180 - azAngs
        elif 'RI' in available_pickup:
            magnet.rotation_pickup = 0
            if azAngs < 90:
                magnet.rotation_putdown = 360 - azAngs
            elif azAngs >= 90:
                magnet.rotation_putdown = 360 - azAngs

    # rotation for pickup and putdown of rectangular magnets
    elif magnet.__class__.__name__ == 'rectangular_magnet':
        magnet.rotation_pickup = 180
        ang_azAngs = convert_radians_to_degrees(magnet.rectangular_magnet_input_orientation) + 270
        if ang_azAngs < 180 and azAngs < 90:
            magnet.rotation_putdown = (90 - azAngs) + (180 - ang_azAngs)
            if 'I' in available_pickup:
                magnet.rotation_putdown = magnet.rotation_putdown + 180
        elif ang_azAngs < 180 and azAngs >= 90:
            magnet.rotation_putdown = (450 - azAngs) - (180 + ang_azAngs)
            if 'I' in available_pickup:
                magnet.rotation_putdown = magnet.rotation_putdown + 180
        elif ang_azAngs > 180 and azAngs < 90:
            magnet.rotation_putdown = (90 - azAngs) - (ang_azAngs - 180)
            if 'I' in available_pickup:
                magnet.rotation_putdown = magnet.rotation_putdown - 180
        elif ang_azAngs > 180 and azAngs >= 90:
            magnet.rotation_putdown = (450 - azAngs) + (180 - ang_azAngs)
            if 'I' in available_pickup:
                magnet.rotation_putdown = magnet.rotation_putdown - 180

    return magnet


# checking angle to make sure it is within the range of 0 to 360
def check_for_negative_twopi_angle(angle):

    if angle < 0:
        angle = 360 + angle
    elif angle > 360:
        angle = angle - 360

    return angle




from ...general_operations.geometry_shapes.segment import segment
from ...general_operations.geometry_intersections.circle_with_segment import create_circle_segment_equation

def circle_rectangle_intersection(circle,rectangle):

    rectangle_corners = rectangle.calculate_4corners()

    all_intersecting_points = []

    for k, l in zip([0, 1, 2, 3], [1, 2, 3, 0]):

            rectangle_side = segment(rectangle_corners[k], rectangle_corners[l])

            equation = create_circle_segment_equation(circle, rectangle_side)
            points = equation.solve_segment_intersects_with_circle()

            if points:
                all_intersecting_points.extend(points)

    return all_intersecting_points



import sys

from ...hector.magnets.pickup_areas import tangential_right, tangential_left, radial_inward,radial_outward,inward,outward

class conflicted_magnet:

    def __init__(self,blocking_magnet,blocked_magnet,blocked_pickup_area):

        self.blocking_magnet     = blocking_magnet
        self.blocked_magnet      = blocked_magnet
        self.blocked_pickup_area = blocked_pickup_area

    def remove_blocked_pickup_areas_from_list_available_pickup(self):

        if self.blocked_pickup_area in self.blocked_magnet.pickup_areas:
            self.blocked_magnet.pickup_areas.remove(self.blocked_pickup_area)

def circular_magnet_is_fully_blocked(list_of_blocked_pickup_areas):

    [blocked_tangential_right,\
     blocked_tangential_left,\
     blocked_radial_inward,\
     blocked_radial_outward] = [False, False, False, False]

    for blocked_pickup_area in list_of_blocked_pickup_areas:

        if isinstance(blocked_pickup_area,tangential_right):
            blocked_tangential_right = True

        if isinstance(blocked_pickup_area,tangential_left):
            blocked_tangential_left = True

        if isinstance(blocked_pickup_area,radial_inward):
            blocked_radial_inward = True

        if isinstance(blocked_pickup_area,radial_outward):
            blocked_radial_outward = True

    if blocked_tangential_right and blocked_tangential_left and blocked_radial_inward and blocked_radial_outward:

        return True

    else:

        return False

def rectangular_magnet_is_fully_blocked(list_of_blocked_pickup_areas):

    [blocked_inward, blocked_outward] = [False, False]

    for pickup_area in list_of_blocked_pickup_areas:

        if isinstance(pickup_area,inward):
            blocked_inward = True

        if isinstance(pickup_area,outward):
            blocked_outward = True

    if blocked_inward and blocked_outward:

        return True

    else:

        return False

def print_fully_blocked_magnets(fully_blocked_magnets, conflictsRecord, fileNameHexa):

    print('Fully blocked magnets:')

    for magnet in fully_blocked_magnets:
        print(magnet.__class__.__name__,int(magnet.index))
        #sys.exit('CONFLICTING MAGNETS FOUND!')

        with open(conflictsRecord,'a') as fp:
            conflict = magnet.__class__.__name__ + ' ' + str(int(magnet.index)) + ' ' + fileNameHexa + '\n'
            fp.write(conflict)

    if fully_blocked_magnets == []:
        print('None.')



from .constants import circular_rectangle_magnet_center_distance,HECTOR_plate_radius
from math import pi, cos, sin, atan
import numpy as np
from ..general_operations.trigonometry import convert_radians_to_degrees, convert_modulus_angle, convert_degrees_to_radians
from .magnets.circular import circular_magnet
from .magnets.rectangular import rectangular_magnet

class probe:

    def __init__(self, probe_index, circular_magnet_center, rectangular_magnet_input_orientation, galaxyORstar, Re, mu_1re, Mstar, magnet_label, hexabundle, rads, rotation_pickup, rotation_putdown, azAngs, IDs):

        self.index = probe_index
        self.circular_magnet_center = np.array(circular_magnet_center) * HECTOR_plate_radius
        self.rectangular_magnet_input_orientation = rectangular_magnet_input_orientation
        self.circular_rectangle_magnet_center_distance = circular_rectangle_magnet_center_distance
        self.galaxyORstar = galaxyORstar
        self.Re = Re
        self.mu_1re = mu_1re
        self.Mstar = Mstar
        self.magnet_label = magnet_label
        self.hexabundle = hexabundle
        self.rads = rads
        self.rotation_pickup = rotation_pickup
        self.rotation_putdown = rotation_putdown
        self.azAngs = azAngs
        self.IDs = IDs

        # self.circular_magnet_center[0] = - (self.circular_magnet_center[0])

    def calculate_circular_magnet_orientation(self):

        # print(self.circular_magnet_center) #test case 2

        if self.circular_magnet_center[0] >= 0 and self.circular_magnet_center[1] >= 0:

            self.circular_magnet_orientation = \
            atan(np.abs(self.circular_magnet_center[1] / self.circular_magnet_center[0]))

        elif self.circular_magnet_center[0] < 0 and self.circular_magnet_center[1] >= 0:

            self.circular_magnet_orientation = pi - \
            atan(np.abs(self.circular_magnet_center[1] / self.circular_magnet_center[0]))

        elif self.circular_magnet_center[0] < 0 and self.circular_magnet_center[1] < 0:

            self.circular_magnet_orientation = \
            pi + atan(np.abs(self.circular_magnet_center[1] / self.circular_magnet_center[0]))

        elif self.circular_magnet_center[0] >= 0 and self.circular_magnet_center[1] < 0:

            self.circular_magnet_orientation = \
            2 * pi - atan(np.abs(self.circular_magnet_center[1] / self.circular_magnet_center[0]))

        # print(self.circular_magnet_orientation)  # test case

        self.circular_magnet_orientation = convert_radians_to_degrees(self.circular_magnet_orientation)

        return self.circular_magnet_orientation

    def calculate_rectangular_magnet_orientation(self):

       probe.calculate_circular_magnet_orientation(self)

       self.rectangular_magnet_absolute_orientation_degree = \
       90 - self.circular_magnet_orientation - convert_radians_to_degrees(self.rectangular_magnet_input_orientation)

       return self.rectangular_magnet_absolute_orientation_degree

    def calculate_rectangular_magnet_center_coordinates(self):

        probe.calculate_rectangular_magnet_orientation(self)

        self.rectangular_magnet_orientation_modulo_radians = \
        convert_modulus_angle(convert_degrees_to_radians(90 - self.rectangular_magnet_absolute_orientation_degree))

        if 0 < self.rectangular_magnet_orientation_modulo_radians < pi / 2:

            self.rectangular_magnet_center = \
            [self.circular_magnet_center[0] \
             + self.circular_rectangle_magnet_center_distance * cos(self.rectangular_magnet_orientation_modulo_radians),
             self.circular_magnet_center[1] \
             + self.circular_rectangle_magnet_center_distance * sin(self.rectangular_magnet_orientation_modulo_radians)]

        elif pi / 2 < self.rectangular_magnet_orientation_modulo_radians < pi:

            self.rectangular_magnet_center = \
            [self.circular_magnet_center[0] \
             - self.circular_rectangle_magnet_center_distance * cos(pi - self.rectangular_magnet_orientation_modulo_radians),
             self.circular_magnet_center[1] \
             + self.circular_rectangle_magnet_center_distance * sin(pi - self.rectangular_magnet_orientation_modulo_radians)]

        elif pi < self.rectangular_magnet_orientation_modulo_radians < 3 * pi / 2:

            self.rectangular_magnet_center = \
            [self.circular_magnet_center[0] \
             - self.circular_rectangle_magnet_center_distance * sin(3 * pi / 2 - self.rectangular_magnet_orientation_modulo_radians),
             self.circular_magnet_center[1] \
             - self.circular_rectangle_magnet_center_distance * cos(3 * pi / 2 - self.rectangular_magnet_orientation_modulo_radians)]

        elif 3 * pi / 2 < self.rectangular_magnet_orientation_modulo_radians < 2 * pi:

            self.rectangular_magnet_center = \
            [self.circular_magnet_center[0] \
             + self.circular_rectangle_magnet_center_distance * cos(2 * pi - self.rectangular_magnet_orientation_modulo_radians),
             self.circular_magnet_center[1] \
             - self.circular_rectangle_magnet_center_distance * sin(2 * pi - self.rectangular_magnet_orientation_modulo_radians)]

        return np.array(self.rectangular_magnet_center)

    def extract_circular_magnet_parameters(self):
        return circular_magnet(self.circular_magnet_center, probe.calculate_circular_magnet_orientation(self), self.index, self.galaxyORstar, self.Re,
                    self.mu_1re, self.Mstar, self.magnet_label, self.hexabundle, self.rads, self.rotation_pickup, self.rotation_putdown, self.azAngs, self.rectangular_magnet_input_orientation, self.IDs)

    def extract_rectangular_magnet_parameters(self):
        return rectangular_magnet(probe.calculate_rectangular_magnet_center_coordinates(self),
                                  probe.calculate_rectangular_magnet_orientation(self),self.index, self.galaxyORstar,self.Re,self.mu_1re,self.Mstar,
                    self.magnet_label, self.hexabundle, self.rads, self.rotation_pickup, self.rotation_putdown, self.azAngs, self.rectangular_magnet_input_orientation, self.IDs)

